import os
from pathlib import Path

import pandas as pd

"""
@article{yang2020causal,
  title={Causal intersectionality for fair ranking},
  author={Yang, Ke and Loftus, Joshua R and Stoyanovich, Julia},
  journal={arXiv preprint arXiv:2006.08688},
  year={2020}
}

License
Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated 
documentation files (the "Software"), to deal in the Software without restriction, including without limitation 
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, 
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE 
OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
project_dir = Path.cwd()
def run_causal_model(data, configs, dataset, model_path):
    """
    Runs the causal model estimation on the data and saves the estimates at self.model_path
    Args:
        data (pandas.Dataframe): Data to be used for estimating the causal effects of the sensitive attributes on the data.
        configs (dict): Configuration dict of the fairness method.
    """
    from rpy2 import robjects
    from rpy2.robjects import pandas2ri

    temp = data[data[dataset.score_col] > dataset.pos_th]
    temp = temp[[dataset.query_col, dataset.ID, configs['group'],
                 dataset.score_col] + dataset.feature_cols]

    try:
        pandas2ri.activate()
        r = robjects.r
        r_script = os.path.join(project_dir, "search", "preprocessing_model", "modules", "CIFRank",
                                "R", "estimate_causal_model.R")


        if not os.path.exists(os.path.join(model_path)):
            os.makedirs(os.path.join(model_path))

        r.source(r_script, encoding="utf-8")
        r.estimate_causal_model(temp, configs['group'], dataset.score_col,
                                dataset.feature_cols, configs['control'],
                                os.path.join(model_path))
    except Exception as e:
        with open(os.path.join(model_path, "logs.txt"), 'a') as f:
            f.write("error in causal model: " + str(e) + "\n")

        if len(os.listdir(model_path)) != 0:
            df = pd.DataFrame(columns=["Mediators"])
            df["Mediators"] = 'nan'
            df.to_csv(os.path.join(model_path, 'identified_mediators.csv'))
    groups = temp[configs['group']].unique()
    save_med_results(groups, configs['control'], os.path.join(model_path))


def save_med_results(groups, control, out_path):
    """Save the output of the mediation analysis.
    Args:
        groups (list(str)): List containing the values of the sensitive column of the data.
        control (str): Control value used in the causal estimation.
        out_path (str): Path to the output file where the mediation analysis is saved.
    """
    if os.path.exists(os.path.join(out_path, 'med_output.txt')):
        with open(os.path.join(out_path, 'med_output.txt'), 'r') as f:
            content = f.readlines()
        results_dict = dict()
        next_indirect = False
        for line in content:
            line = line.strip()
            if line.startswith('For the predictor'):
                if len(results_dict.keys()) == 0:
                    pred = line.split(' ')[3]
                    df_med = pd.DataFrame(columns=['Metric', 'Estimate'])
                    results_dict[pred] = ''
                else:
                    results_dict[pred] = df_med
                    pred = line.split(' ')[3]
                    df_med = pd.DataFrame(columns=['Metric', 'Estimate'])

            if line.startswith('The estimated total effect:'):
                total_effect = float(line.split(' ')[4])
                temp_df = pd.DataFrame([['Total Effect', total_effect]], columns=['Metric', 'Estimate'])
                df_med = pd.concat([df_med, temp_df], ignore_index=True)

            if next_indirect:
                splits = line.split(' ')
                if splits[0] == '':
                    indirect_effect = float(line.split(' ')[1])
                else:
                    indirect_effect = float(line.split(' ')[0])
                temp_df = pd.DataFrame([['Indirect Effect', indirect_effect]], columns=['Metric', 'Estimate'])
                df_med = pd.concat([df_med, temp_df], ignore_index=True)
                next_indirect = False

            if line.startswith('y1.all'):
                next_indirect = True

        results_dict[pred] = df_med

        pred_groups = [p.split('pred')[1] for p in results_dict.keys()]
        pred_gr = [g for g in groups if g not in pred_groups and g != control][0]

        index = 0
        for key in results_dict.keys():
            index = index + 1
            df_med = results_dict[key]
            direct_effect = df_med[df_med['Metric'] == 'Total Effect']['Estimate'].values[0] - \
                            df_med[df_med['Metric'] == 'Indirect Effect']['Estimate'].values[0]
            temp_df = pd.DataFrame([['Direct Effect', direct_effect]], columns=['Metric', 'Estimate'])
            df_med = pd.concat([df_med, temp_df], ignore_index=True)

            if key == 'pred' or key == '':
                file_name = pred_gr + '_med.csv'
            elif 'pred.temp1$x' in key:
                file_name = groups[index] + '_med.csv'
            else:
                file_name = key.split('pred')[1] + '_med.csv'

            df_med.to_csv(os.path.join(out_path, file_name))


import os
import numpy as np
import json
from scipy.sparse import coo_matrix, csr_matrix
from ..utils import Build_Adjecent_Matrix


class Abstract_Reranker(object):
    def __init__(self, config, weights = None):
        self.config = config
        self.item_num = config['item_num']
        self.group_num = config['group_num']
        if not weights:
            weights = np.ones(self.group_num)
        self.weights = weights
        self.M, self.iid2pid = Build_Adjecent_Matrix(config)


    def rerank(self, ranking_score, k):
        """
           Re-ranks the items based on the initial ranking scores and a fairness regularization term.

           This function performs re-ranking of items for each user by incorporating a fairness regularization term
           (`minimax_reg`) that adjusts the ranking scores to promote fairness across groups. The re-ranked list of
           items is returned for each user.

           :param ranking_score: A 2D array (or tensor) of ranking scores for all items, with shape (user_size, item_num).
                                  Each row corresponds to the scores for a user and each column corresponds to an item.
           :param k: The number of top-ranked items to return for each user.
           :return: A list of re-ranked item indices for each user, with the top `k` items for each user.
        """
        pass



import os
import numpy as np
from .Abstract_Reranker import Abstract_Reranker
from tqdm import tqdm,trange

r"""
K-neighbor: it is a heuristic baseline for re-ranking: every user arrives, it will only choose the least-k worst-off group's items to recommend
"""

class k_neighbor(Abstract_Reranker):
    def __init__(self, config, weights = None):
        super().__init__(config, weights)


    def rerank(self, ranking_score, k):
        ## its parameters
        user_size = len(ranking_score)
        B_l = np.zeros(self.group_num)
        rerank_list = []

        for u in trange(user_size):
            worstoff_groups = np.argsort(B_l)
            worstoff_groups = worstoff_groups[:self.config['nearest_k']]
            target_onehot_group = np.zeros(self.config['group_num'])
            target_onehot_group[worstoff_groups] = 1
            target_items = np.matmul(self.M, target_onehot_group)
            recommended_mask = (1 - target_items) * -10000.0
            rel = ranking_score[u,:]
            result_item = np.argsort(rel+recommended_mask)[::-1]
            result_item = result_item[:k]
            rerank_list.append(result_item)
            B_l = B_l + np.sum(self.M[result_item,:],axis=0,keepdims=False)


        return rerank_list



import os
import torch
import pickle
import random
import numpy as np
import multiprocessing

from tqdm import tqdm

from .utils import load_embedding, get_rel_feat, pkl_save, pkl_load, split_list


MAXDOC=50
MAX_DIV_DIM = 8
REL_LEN=18

def build_each_train_dataset(qid_list, qd, train_dict, rel_feat_dict, res_dir, query_emb, doc_emb):
    """
    Generates and saves the training dataset for each query in qid_list. 
    
    :param qid_list: A list of query IDs to process.
    :param qd: A dictionary containing the query data (query, document list, subtopic info).
    :param train_dict: A dictionary of pre-generated training samples.
    :param rel_feat_dict: A dictionary of relevance features for query-document pairs.
    :param res_dir: The directory where the processed data will be saved.
    :param query_emb: A dictionary of query embeddings.
    :param doc_emb: A dictionary of document embeddings.
    :return: Saves the processed data for each query in a .pkl.gz file. 
    """

    for qid in tqdm(qid_list, desc="GenTrainData", ncols = 90):
        sample_dict={}
        sample_path = res_dir + str(qid) + '.pkl.gz'

        sample_feat_list = []
        sample_rel_list = []
        sample_div_list = []
        query = qd[str(qid)].query
        Doc_list = qd[str(qid)].doc_list
        df = np.array(qd[qid].subtopic_df)
        sample_list = train_dict[str(qid)]
        count = -1
        while count < len(sample_list):
            if count == -1:
                doc_list = qd[str(qid)].best_docs_rank
            else:
                doc_list = sample_list[count]
            
            rel_feat_list = []
            div_labels = []
            feat_list = []
            feat_list.append(torch.tensor(query_emb[query]).float())

            for i in range(len(doc_list)):
                rel_feat = rel_feat_dict[(query, doc_list[i])]
                rel_feat_list.append(torch.tensor(rel_feat).float())
                doc_feat = doc_emb[doc_list[i]]
                feat_list.append(torch.tensor(doc_feat).float())
                index = Doc_list.index(doc_list[i])
                div_feat = list(df[index, :])
                if len(div_feat) < MAX_DIV_DIM:
                    div_feat.extend([0]*(MAX_DIV_DIM - len(div_feat)))
                div_labels.append(torch.tensor(div_feat).float())
            
            if len(feat_list) < (MAXDOC+1):
                feat_list.extend([torch.tensor([0]*100).float()]*(MAXDOC+1-len(feat_list)))
            if len(div_labels) < MAXDOC:
                div_labels.extend([torch.tensor([0]*MAX_DIV_DIM).float()]*(MAXDOC-len(div_labels)))
            if len(rel_feat_list) < MAXDOC:
                rel_feat_list.extend([torch.tensor([0]*REL_LEN).float()]*(MAXDOC-len(rel_feat_list)))
            
            feat_tensor = torch.stack(feat_list, dim=0).float()
            rel_feat = torch.stack(rel_feat_list, dim=0).float()
            div_tensor = torch.stack(div_labels, dim=0).float()

            assert feat_tensor.shape[0] == (MAXDOC+1)
            assert rel_feat.shape[0] == (MAXDOC)
            assert div_tensor.shape[0] == MAXDOC
            if div_tensor.shape[1] != MAX_DIV_DIM:
                print('qid = {}, len={}'.format(qid, div_tensor.shape[1]))
            assert div_tensor.shape[1] == MAX_DIV_DIM
            sample_feat_list.append(feat_tensor)
            sample_rel_list.append(rel_feat)
            sample_div_list.append(div_tensor)
            count += 1
        
        assert len(sample_feat_list) == len(sample_div_list)
        assert len(sample_rel_list) == len(sample_div_list)
        sample_dict[qid]=[
            (sample_feat_list[i],
            sample_rel_list[i],
            sample_div_list[i])
            for i in range(len(sample_feat_list))
        ]
        pkl_save(sample_dict, sample_path)


def build_train_dataset(config, worker_num=20):
    """
    Builds the training dataset by distributing the workload across multiple workers. 
    
    :param config: A dictionary containing configuration settings.
    :param worker_num: The number of workers to use for parallel processing.
    :return: Generates the training dataset and saves it in the result directory.
    """
    res_dir=os.path.join(config['data_dir'], config['model'], 'train/')
    if not os.path.exists(res_dir):
        os.makedirs(res_dir)
    all_qids=np.load(os.path.join(config['data_dir'], 'all_qids.npy'))
    qd = pickle.load(open(os.path.join(config['data_dir'], 'div_query.data'), 'rb'))
    doc_emb = load_embedding(os.path.join(config['data_dir'], config['embedding_dir'], config['embedding_type']+'_doc.emb'))
    query_emb = load_embedding(os.path.join(config['data_dir'], config['embedding_dir'], config['embedding_type']+'_query.emb'))
    train_dict = pkl_load(os.path.join(config['data_dir'], config['model'], 'list_train_samples.pkl.gz'))
    rel_feat_dict = get_rel_feat(os.path.join(config['data_dir'], 'rel_feat.csv'))

    task_list = split_list(all_qids, worker_num)
    jobs=[]
    for task in task_list:
        p = multiprocessing.Process(target=build_each_train_dataset, args=(task, qd, train_dict, rel_feat_dict, res_dir, query_emb, doc_emb))
        jobs.append(p)
        p.start()
    for job in jobs:
        job.join()


def build_test_dataset(config):
    """
    Builds the test dataset for evaluation. 
    
    :param config: A dictionary containing configuration settings.
    :return: Generates the test dataset and saves it in the result directory.
    """
    qd = pickle.load(open(os.path.join(config['data_dir'], 'div_query.data'), 'rb'))
    doc_emb = load_embedding(os.path.join(config['data_dir'], config['embedding_dir'], config['embedding_type']+'_doc.emb'))
    query_emb = load_embedding(os.path.join(config['data_dir'], config['embedding_dir'], config['embedding_type']+'_query.emb'))
    output_dir = os.path.join(config['data_dir'], config['model'], 'test/')
    rel_feat_dict = get_rel_feat(os.path.join(config['data_dir'], 'rel_feat.csv'))
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)
    all_qids=range(1,201)
    del_index=[94,99]
    all_qids=np.delete(all_qids,del_index)
    qids=[str(i) for i in all_qids]
    for qid in tqdm(qids,desc="gen Test", ncols=80):
        print('qid=',qid)
        test_dict = {}
        query = qd[str(qid)].query
        output_file_path = output_dir + str(qid) + '.pkl.gz'
        doc_list = qd[str(qid)].doc_list[:50]
        real_num = len(doc_list)
        
        feat_list = []
        rel_feat_list = []

        feat_list.append(torch.tensor(query_emb[query]).float())

        for i in range(len(doc_list)):
            doc_feat = doc_emb[doc_list[i]]
            feat_list.append(torch.tensor(doc_feat).float())
            rel_feat = torch.tensor(rel_feat_dict[(query, doc_list[i])]).float()
            rel_feat_list.append(rel_feat)

        if len(feat_list) < (MAXDOC+1):
            feat_list.extend([torch.tensor([0]*100).float()]*(MAXDOC+1-len(feat_list)))
        if len(rel_feat_list) < MAXDOC:
            rel_feat_list.extend([torch.tensor([0]*REL_LEN).float()]*(MAXDOC-len(rel_feat_list)))
        
        feat_tensor = torch.stack(feat_list, dim=0).float()
        rel_feat_tensor = torch.stack(rel_feat_list, dim=0).float()
        assert feat_tensor.shape[0] == (MAXDOC+1)
        assert rel_feat_tensor.shape[0] == MAXDOC

        test_dict[qid]=(
            feat_tensor,
            rel_feat_tensor
        )
        pkl_save(test_dict, output_file_path)


def gen_list_training_sample(config, top_n = 50, sample_num = 200):
    """
    Generates list training samples by selecting top-ranked documents for each query. 

    :param config: A dictionary containing configuration settings.
    :param top_n: The number of top-ranked documents to consider for each sample.
    :param sample_num: The number of samples to generate for each query.
    :return: Saves the generated training samples in a file for later use.
    """
    qd = pickle.load(open(os.path.join(config['data_dir'], 'div_query.data'),'rb'))
    doc_emb = load_embedding(os.path.join(config['data_dir'], config['embedding_dir'], config['embedding_type']+'_doc.emb'))
    rel_feat_dict = get_rel_feat(os.path.join(config['data_dir'], 'rel_feat.csv'))
    train_dict={}
    for qid in tqdm(qd, desc="Gen Train"):
        temp_q=qd[qid]
        temp_doc_list = temp_q.doc_list[:100]
        result_list=[]
        real_num=int(min(top_n, temp_q.DOC_NUM))
        for i in range(sample_num):
            random.shuffle(temp_doc_list)
            top_docs = temp_doc_list[:real_num]
            flag = 0
            for j in range(len(top_docs)):
                if (qid, top_docs[j]) not in rel_feat_dict:
                    flag = 1
                    break
            if flag == 0 and top_docs not in result_list:
                result_list.append(top_docs)
        print('qid={}, len={}'.format(qid, len(result_list)))
        train_dict[str(qid)]=result_list
    if not os.path.exists(os.path.join(config['data_dir'], config['model'])):
            os.makedirs(os.path.join(config['data_dir'], config['model']))
    pkl_save(train_dict, os.path.join(config['data_dir'], config['model'], 'list_train_samples.pkl.gz'))


def Process(config):
    """ Main function for DALETOR data processing.
    
    :param config: A dictionary containing configuration settings.
    :return: Generates and saves the training and test datasets, as well as list training samples.
    """
    gen_list_training_sample(config)
    build_train_dataset(config)
    build_test_dataset(config)


import os.path
from .utils.group_utils import Init_Group_AdjcentMatrix, get_iid2text
import pandas as pd
from .llm import Grounder, Prompt_Constructer, LLM_caller
from .evaluator import LLM_Evaluator
import os
import yaml
import json
from datetime import datetime

class LLMRecommender(object):
    """

    """
    def __init__(self, train_config):
        """Initialize In-processing and base LLMs-models.

            :param train_config: Your custom config files.
        """
        self.dataset = train_config['dataset']
        #self.stage = stage
        self.llm_type = train_config['llm_type']
        self.config = train_config

    def load_configs(self, dir):
        """
            Loads and merges configuration files for the model, dataset, and evaluation.

            This function loads multiple YAML configuration files, including the process configuration,
            dataset-specific settings, model configurations, and evaluation parameters. All configurations
            are merged, with the highest priority given to the class's own `config` attribute.

            :param dir: The directory where the main process configuration file is located.
            :return: A dictionary containing the merged configuration from all files.
        """
        print("start to load config...")
        with open(os.path.join(dir, "process_config.yaml"), 'r') as f:
            config = yaml.safe_load(f)

        with open(os.path.join("recommendation", "properties", "dataset", f"{self.dataset}.yaml"),
                  'r') as f:
            config.update(yaml.safe_load(f))

        print("start to load model...")
        with open(os.path.join("recommendation", "properties", "models.yaml"), 'r') as f:
            model_config = yaml.safe_load(f)

        with open(os.path.join("recommendation", "properties", "models", f"LLM.yaml"),
                  'r') as f:
            model_config.update(yaml.safe_load(f))

        config.update(model_config)

        with open(os.path.join("recommendation", "properties", "evaluation.yaml"), 'r') as f:
            config.update(yaml.safe_load(f))

        config.update(self.config)  ###train_config has highest rights
        print("your loading config is:")
        print(config)

        return config

    def recommend(self):
        """
            Training LLMs-based in-processing and base model main workflow.
        """

        dataset_dir = os.path.join("recommendation", "processed_dataset", self.dataset)
        dataset_file_name = self.dataset + '.test.ranking'
        input_file = pd.read_csv(os.path.join(dataset_dir, dataset_file_name), delimiter='\t')
        iid2text, iid2pid = get_iid2text(self.dataset), Init_Group_AdjcentMatrix(self.dataset)
        config = self.load_configs(dataset_dir)
        prompt_constructer = Prompt_Constructer(config)
        prompt_dataset = prompt_constructer.construct_prompt(input_file, iid2text, iid2pid)

        LLM = LLM_caller(config)

        results_list = LLM.get_response(prompt_dataset)
        LLM.clear()
        grounder = Grounder(config)

        grounding_result = grounder.grounding(results_list, id2title=iid2text)

        evaluator = LLM_Evaluator(config)

        eval_result = evaluator.llm_eval(grounding_result, iid2pid)

        # dump the logs and eval result
        today = datetime.today()
        today_str = f"{today.year}-{today.month}-{today.day}"
        log_dir = os.path.join("recommendation", "log", f"{today_str}_{config['log_name']}")
        if not os.path.exists(log_dir):
            os.makedirs(log_dir)


        print(f"training complete! start to save the config and model...")
        print(f" config files are dump in {log_dir}")
        with open(os.path.join(log_dir, "config.yaml"), 'w') as f:
            yaml.dump(config, f)


        with open(os.path.join(log_dir, 'test_result.json'), 'w') as f:
            json.dump(eval_result, f)
        print(f"dump in {log_dir}")





import numpy as np
from .Abstract_Ranker import Abstract_Reweigher

r"""
Inverse Propensity Scores Re-weighter
################################################

The heuristic baselines, which utilizes the item popularity weight as the inverse propensity scores for the different samples
Item-side fairness of large language model-based recommendation system, WWW Web4Good 2024, Jiang et al.

"""

class IPS(Abstract_Reweigher):
    def __init__(self, config, group_weight):
        super().__init__(config)
        self.variance_control = config['M']
        self.group_weight = group_weight


    def reweight(self, input_dict):
        items = input_dict['items']

        adj_matrix = self.M[items]

        B_t = np.sum(adj_matrix, axis=0, keepdims=False)
        self.exposure_count = self.exposure_count + B_t
        norm_count = self.group_weight * self.exposure_count / np.sum(self.exposure_count)
        batch_weight = np.matmul(adj_matrix, norm_count)
        batch_weight = batch_weight / np.sum(batch_weight)

        return 1/(batch_weight+self.variance_control)




import os
import pandas as pd
import ast
from ..utils.process_tabular_data import create_train_split


class Compas:
    """
    A class to handle the COMPAS dataset, process it, and create train-test splits.
    """

    def __init__(self, config):
        """
        Initializes the COMPAS dataset handler.

        :param config: Dictionary containing dataset configuration parameters.
        """
        self.path = config["path"]
        self.data_path = os.path.join(config["path"], "COMPAS.csv")  # Dataset file path
        self.query_col = config["query_col"]  # Query column name
        self.ID = config["id_col"]  # Unique identifier column
        self.sensitive_cols = config["sensitive_cols"]  # Sensitive attribute columns
        self.score_col = config["score_col"]  # Score column
        self.feature_cols = config["feature_cols"]  # Feature columns
        self.k_fold = config["k_fold"]  # Number of folds for cross-validation
        self.ratio_split = config["ratio_split"]  # Train-test split ratio
        self.pos_th = config["pos_th"]  # Positive threshold for classification

        self.processed_file = os.path.join(self.path, "COMPAS_processed.csv")

        # Process the data during initialization
        self.process_data()

    def process_data(self):
        """
        Processes the dataset and creates train-test splits.
        """

        if not os.path.exists(self.processed_file):
            data = pd.read_csv(self.data_path)

            # Assign the same query ID to all rows (since there are no queries)
            data[self.query_col] = 0

            # Handle intersectional sensitive columns if applicable
            if "__" in self.sensitive_cols:
                intersectional_column = self.sensitive_cols[0]
                groups = intersectional_column.split("__")
                data[self.sensitive_cols] = data[groups].agg("__".join, axis=1)

            # Select relevant columns
            selected_columns = [self.query_col, self.ID, *self.feature_cols, self.score_col, *self.sensitive_cols]
            data = data[selected_columns]

            # Save the processed dataset
            data.to_csv(self.processed_file, index=False)

            # Create train-test splits
            split_path = os.path.join(self.path, "splits_data")
            os.makedirs(split_path, exist_ok=True)
            create_train_split(data, self.query_col, self.ID, self.sensitive_cols, self.score_col,
                               self.pos_th, split_path, self.k_fold, 1 - self.ratio_split)

    def read_data(self, i):
        """
        Reads the ith fold for train-test splits.

        :param i: The fold index (for k-fold cross-validation).
        :return: Tuple (data_train:Pandas DataFrame, data_test:Pandas DataFrames).
        """
        data = pd.read_csv(self.processed_file)

        # Paths to train-test split files
        split_path = os.path.join(self.path, "splits_data")
        train_file = os.path.join(split_path, f'train_samples_{i}.txt')
        test_file = os.path.join(split_path, f'test_samples_{i}.txt')

        # Load train-test IDs
        train_ids = self._load_ids(train_file)
        test_ids = self._load_ids(test_file)

        # Split dataset
        data_train = data[data[self.ID].astype(str).isin(train_ids)]
        data_test = data[data[self.ID].astype(str).isin(test_ids)]

        return data_train, data_test

    @staticmethod
    def _load_ids(file_path):
        """
        Helper method to read IDs from a text file.

        :param file_path: Path to the file containing IDs.
        :return: List of IDs.
        """
        with open(file_path, 'r') as f:
            return f.read().splitlines()



import os
from abc import ABC, abstractmethod
from pathlib import Path

class PreprocessingFairnessIntervention(ABC):
    """
        Abstract base class for preprocessing fairness interventions.

        This class serves as a base for various fairness intervention methods that preprocess datasets
        to reduce biases while maintaining data utility.
    """
    def __init__(self, configs, dataset):
        """
            Initialize the fairness intervention with the given configurations and dataset.

            This method sets up the necessary directories for storing model parameters and
            fair data transformations.

            :param configs : dict
                Configuration dictionary containing model parameters and paths.
            :param dataset : object
                Dataset object containing path information.

            Attributes:
            - self.dataset : object
                Stores the dataset object for reference.
            - self.configs : dict
                Stores the configuration parameters.
            - self.model_path : Path
                Directory path where model parameters will be stored.
            - self.fair_data_path : Path
                Directory path where transformed fair data will be stored.

            Note:
            - The necessary directories are automatically created if they do not exist.
        """
        self.dataset = dataset
        self.configs = configs
        self.model_path = Path(self.dataset.path) / self.configs["name"] / "model_params"
        os.makedirs(self.model_path, exist_ok=True)
        self.fair_data_path = Path(self.dataset.path) / self.configs["name"] / "fair_data"
        os.makedirs(self.fair_data_path, exist_ok=True)
    @abstractmethod
    def fit(self, X_train, run):
        """
        Train the fairness model using the given training dataset.

        :param X_train : pandas.DataFrame
            The training dataset.
        :param run : str
            The identifier for the training run when using k-fold.

        :return : None
        """
        pass

    @abstractmethod
    def transform(self, X, run, file_name=None):
        """
            Apply the transformation to the dataset using the learned model.

            :param X : pandas.DataFrame
                The dataset to which the fairness transformation is applied.
            :param run : str
                The identifier for the transformation run.
            :param file_name : str, optional
                Name of the file to save the transformed dataset.

            :return : pandas.DataFrame
                The dataset with transformed fair columns.
        """
        pass

    def fit_transform(self, X_train, run):
        """
        Learns the model from the training data and returns the data in the new space.
        """
        print('Fitting and transforming...')
        self.fit(X_train, run)
        return self.transform(X_train, run)


import os
import re
import json
import numpy as np
import pandas as pd
import torch as th
import pickle
import torch
from .postprocessing_model.DESA import DESA
from .postprocessing_model.DALETOR import DALETOR
from .utils.utils import get_metrics_20

MAXDOC = 50
REL_LEN = 18


def evaluate_test_qids_DESA(model, test_tuple, div_q, mode='metric'):
    """
    Get the alpha-nDCG for the input query, the input document list are randomly shuffled.
    :param test_tuple: the features of the test query qid, test_turple = {}
    :param div_q: the div_query object of the test query qid
    :param qid: the id for the test query
    :return: the alpha-nDCG for the test query
    """

    metric = 0
    end = Max_doc_num = len(div_q.best_docs_rank)
    current_docs_rank = []
    if not test_tuple:
        return 0
    else:
        doc_mask = test_tuple['doc2vec_mask'].unsqueeze(0) # [1,50]
        sub_mask = test_tuple['sub2vec_mask'].unsqueeze(0) # [1,10]
        doc_emb = test_tuple['doc2vec'].unsqueeze(0).float() # [1, 50, 100]
        sub_emb = test_tuple['sub2vec'].unsqueeze(0).float() # [1,10,100]
        pos_qrel_feat = test_tuple['pos_qrel_feat'].unsqueeze(0).float() # [1,50,18]
        subrel_feat_mask = test_tuple['subrel_feat_mask'].unsqueeze(0)
        pos_subrel_feat = test_tuple['pos_subrel_feat'].unsqueeze(0).float() # [1,50,10,18]

        doc_mask.requires_grad = False
        sub_mask.requires_grad = False
        doc_emb.requires_grad = False
        sub_emb.requires_grad = False
        pos_qrel_feat.requires_grad = False
        subrel_feat_mask.requires_grad = False
        pos_subrel_feat.requires_grad = False

        if th.cuda.is_available():
            doc_mask, sub_mask, doc_emb, sub_emb, pos_qrel_feat, subrel_feat_mask, pos_subrel_feat =\
                doc_mask.cuda(), sub_mask.cuda(), doc_emb.cuda(), sub_emb.cuda(), pos_qrel_feat.cuda(), \
                subrel_feat_mask.cuda(), pos_subrel_feat.cuda()
        #print(doc_emb.shape, sub_emb.shape, doc_mask.shape, sub_mask.shape, pos_qrel_feat.shape, pos_subrel_feat.shape)
        score = model.fit(doc_emb, sub_emb, doc_mask, sub_mask, pos_qrel_feat, pos_subrel_feat, mode='Test')
        result = list(np.argsort(score[:len(test_tuple['doclist'])].cpu().detach().numpy()))
        if len(result) > 0:
            new_docs_rank = []
            for i in range(len(result)-1, -1, -1):
                new_docs_rank.append(test_tuple['doclist'][result[i]])
            #new_docs_rank = [test_tuple['doclist'][result[i]] for i in range(len(result)-1, len(result)-len(test_tuple['doclist'])-1, -1)]
            metric = div_q.get_test_alpha_nDCG(new_docs_rank)
    if mode == 'metric':
        return metric
    elif mode == 'both':
        return metric, new_docs_rank


def evaluate_test_qids_DALETOR(model, test_tuple, div_q, mode='metric'):
    """
    Evaluates the model on a set of test queries and returns the evaluation metric.
    
    :param model: The trained model to be evaluated.
    :param test_tuple: A tuple containing two elements - 
                       the first element is the input features (X),
                       and the second element is the relevance features (rel_feat).
    :param div_q: A `div_query` object that contains information about documents and subtopics.
    :param mode: The evaluation mode. It can be 'metric' (return only metric) or 'both' (return metric and ranked documents).
    :return: The evaluation metric (alpha-nDCG) or a tuple of the metric and the ranked documents, depending on the mode.
    """
    metric = 0
    end = Max_doc_num = len(div_q.best_docs_rank)
    current_docs_rank = []
    if test_tuple[0].shape[0] == 0:
        if mode == 'metric':
            return 0 
        else:
            return []
    else:
        X = test_tuple[0]
        rel_feat = test_tuple[1]
        X.requires_grad = False
        rel_feat.requires_grad = False
        X = X.reshape(1, X.shape[0], X.shape[1])
        rel_feat = rel_feat.reshape(1, rel_feat.shape[0], rel_feat.shape[1])
        
        if th.cuda.is_available():
            X = X.cuda()
            rel_feat = rel_feat.cuda()
        
        outputs = model.fit(X, rel_feat, False)
        out = outputs.cpu().detach().numpy().reshape(MAXDOC)
        # print('out.shape = ',out.shape)
        # print('out = ', out)
        result = np.argsort(-out[:end])
        # print('result =', result)

        for i in range(len(result)):
            if result[i] < Max_doc_num and result[i] not in current_docs_rank:
                current_docs_rank.append(result[i])

        if len(current_docs_rank)>0:
            new_docs_rank = [div_q.doc_list[i] for i in current_docs_rank]
            metric = div_q.get_test_alpha_nDCG(new_docs_rank)
            # print('qid = {}, metric = {}, mode = {}'.format(qid, metric, mode))
            if mode == 'metric':
                return metric
            elif mode == 'both':
                return metric, new_docs_rank


def get_global_fullset_metric(config):
    """
    Get the final metrics for the five fold best models.
    :param best_model_list: the best models for the five corresponding folds.
    :param test_qids_list: the corresponding test qids for five folds.
    """
    output_file = os.path.join(config['tmp_dir'], config['model'], 'run')
    if not os.path.exists(os.path.join(config['tmp_dir'], config['model'])):
        os.makedirs(os.path.join(config['tmp_dir'], config['model']))
    fout = open(output_file, 'w')
    all_models = config['best_model_list']
    qd = pickle.load(open(os.path.join(config['data_dir'], 'div_query.data'), 'rb'))
    with open(os.path.join(config['model_save_dir'], config['model'], 'fold_qid.json'), 'r') as f:
        test_qids_list = json.load(f)

    ''' get the metrics for five folds '''
    fold_time_pattern = re.compile(r'_FOLD_(\d+)_')
    for i in range(len(all_models)):
        model_file = all_models[i]
        fold_times = int(fold_time_pattern.search(all_models[i]).group(1))
        if config['model'].upper() == 'DESA':
            fold_p = os.path.join(config['data_dir'], config['model'], 'fold/')
            test_qids = test_dataset_dict = torch.load(os.path.join(fold_p, 'fold'+str(fold_times), 'test_data.pkl'))
            model = DESA(config['embedding_length'], 8, 2,config['embedding_length'], 8, 2, 8, config['dropout'])
            eval_func = evaluate_test_qids_DESA
        elif config['model'].upper() == 'DALETOR':
            from .datasets.DALETOR import get_test_dataset
            test_qids=test_qids_list[fold_times-1]
            test_dataset_dict = get_test_dataset(i+1, test_qids, config)
            model = DALETOR(0.0)
            eval_func = evaluate_test_qids_DALETOR
        
        model.load_state_dict(th.load(model_file))

        model.eval()
        if th.cuda.is_available():
            model = model.cuda()

        ''' ndeval test '''
        for qid in test_qids:
            metric, docs_rank = eval_func(model, test_dataset_dict[str(qid)], qd[str(qid)], 'both')
            if len(docs_rank)>0:
                for index in range(len(docs_rank)):
                    content = str(qid) + ' Q0 ' + str(docs_rank[index]) + ' ' + str(index+1) + ' -4.04239 indri\n'
                    fout.write(content)
    fout.close()
    csv_path = os.path.join(config['tmp_dir'], config['model'], 'result.csv')
    command = './search/eval/clueweb09/ndeval ./search/eval/clueweb09/2009-2012.diversity.ndeval.qrels ' + output_file + ' >' + str(csv_path)
    os.system(command)
    
    alpha_nDCG_20, NRBP_20, ERR_IA_20, S_rec_20 = get_metrics_20(csv_path)
    print('alpha_nDCG@20_std = {}, NRBP_20 = {}, ERR_IA_20 = {}, S_rec_20 = {}'.format(alpha_nDCG_20, NRBP_20, ERR_IA_20, S_rec_20))




from search.postprocessing_model.xQuAD import xQuAD
from search.postprocessing_model.DESA import DESA
from search.postprocessing_model.PM2 import PM2
from search.postprocessing_model.DALETOR import DALETOR
from search.postprocessing_model.base import BasePostProcessModel, BasePostProcessUnsupervisedModel

diversity_method_mapping = {
    'DESA': DESA,
    'DALETOR': DALETOR,
    'PM2': PM2,
    'xQuAD': xQuAD
}


import os
import time
import json
import pickle
import logging
import argparse
import json_repair

from tqdm import tqdm

from ..llm_model.api_llm import LMAgent
from ..utils.utils import restore_doc_ids, remove_duplicate, get_metrics_20


MAXDOC = 50

adhoc_rerank_prompt_input = """## Input Data
<Query>: {query}
<Document List>: {document_list}

## Output Data
<Output>: 
"""


def llm_run(config):
    """
    Executes a large language model-based document reranking pipeline for search result diversification.
    This function processes queries and their candidate documents through a language model to generate diversified document rankings. 
    
    :param config: Dictionary containing configuration parameters
    """

    best_rank_dir = os.path.join(config['data_dir'], 'best_rank/')
    data_content_dir = os.path.join(config['data_dir'], config['data_content_dir'])

    output_dir = os.path.join(config['tmp_dir'], config['model'], config['model_name'])
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)

    output_txt = os.path.join(output_dir, 'output_best.txt')
    final_output_path = os.path.join(output_dir, 'output_best_new.txt')
    csv_path = os.path.join(output_dir, 'result.csv')
    
    qid_need_process = sorted(os.listdir(best_rank_dir), key=lambda x: int(x.split('.')[0]))
    doc_content_dict = {}
    print("LOAD Document Content ...")
    for fname in os.listdir(data_content_dir):
        with open(data_content_dir+fname, 'r') as fr:
            data_dict = json.load(fr)
            doc_content_dict[fname[:-5]] = data_dict

    adhoc_rerank_prompt_fr = open(os.path.join(config['data_dir'], config['prompts_dir']), 'r')
    adhoc_rerank_prompt = adhoc_rerank_prompt_fr.read()
    adhoc_rerank_prompt_fr.close()

    if not os.path.exists(output_dir):
        os.makedirs(output_dir)

    llm = LMAgent(config)

    for file_name in tqdm(qid_need_process):
        
        qd = pickle.load(open(best_rank_dir+file_name, 'rb'))
        qid = qd.qid
        query = qd.query
        doc_list = qd.doc_list[:MAXDOC]

        print('=====> Query {}. Processing Begin'.format(qid))
        
        docid2num = {doc_id: idx + 1 for idx, doc_id in enumerate(doc_list)}
        document_content_list = '\n'.join(['['+str(docid2num[doc_id])+'] '+' '.join(doc_content_dict[doc_id]['body_text'].split(' ')[:100]) for doc_id in doc_list])
        _adhoc_rerank_prompt = adhoc_rerank_prompt+adhoc_rerank_prompt_input.format(query=query, document_list=document_content_list)
        output = llm('', _adhoc_rerank_prompt, max_new_tokens=config['max_new_tokens'])
        output = output[:output.find('}')+1]
        output = json_repair.loads(output)["rerank_list"]
        doc_ranking = restore_doc_ids(output, docid2num)

        if len(doc_ranking) < 20:
            remain_doc = [item for item in doc_list if item not in doc_ranking]
            doc_ranking = doc_ranking + remain_doc[:20-len(doc_ranking)]
        print('=====> Document Ranking {}. '.format(doc_ranking))

        judge_f = open(os.path.join(output_dir, file_name[:-5]+'.txt'), 'w')
        for i in range(len(doc_ranking)):
            doc_ranking_score = len(doc_ranking) - i
            judge_f.write(str(qid) + ' Q0 ' + doc_ranking[i] + ' ' + str(i + 1) + ' ' + str(doc_ranking_score) + ' indri\n')
        judge_f.close()

    command = 'cat '+output_dir+'/* > '+output_txt
    os.system(command)
    remove_duplicate(output_txt, final_output_path)
    command = './search/eval/clueweb09/ndeval ./search/eval/clueweb09/2009-2012.diversity.ndeval.qrels '+final_output_path+' >' + str(csv_path)
    os.system(command)

    alpha_nDCG_20, NRBP_20, ERR_IA_20, S_rec_20 = get_metrics_20(csv_path)
    print('alpha_nDCG@20_std = {}, NRBP_20 = {}, ERR_IA_20 = {}, S_rec_20 = {}'.format(alpha_nDCG_20, NRBP_20, ERR_IA_20, S_rec_20))

    





import torch
import torch.nn as nn
import numpy as np
from torch.nn.init import normal_

class AbstractBaseModel(nn.Module):
    def __init__(self, config):
        super(AbstractBaseModel, self).__init__()

        self.config = config

        self.user_embedding = nn.Embedding(config['user_num'], config['embedding_size'])
        self.item_embedding = nn.Embedding(config['item_num'], config['embedding_size'])
        self.sigmoid = nn.Sigmoid()

    #def interaction2user(self,interaction):

    def forward(self, **kwargs):
        pass
        # user_embeds = self.user_embedding(user_ids)
        # item_embeds = self.item_embedding(item_ids)
        # dot_product = (user_embeds * item_embeds).sum(1)
        # return dot_product

    def compute_loss(self, **kwargs):
        pass

    def get_user_embedding(self, user):
        return self.user_embedding(user)

    def full_predict(self, user_dict, items):
        # here we assume only one user arrives

        #user = torch.unsqueeze(user)
        #items = torch.unsqueeze(items, 0)

        user_embeds = self.get_user_embedding(user_dict)
        item_embeds = self.item_embedding(items)
        scores = (user_embeds * item_embeds).sum(-1)
        return self.sigmoid(scores)

    def full_ranking(self, user_dict, items, k):
        user_embeds = self.get_user_embedding(user_dict) #[B, D]
        item_embeds = self.item_embedding(items)  # [H, D]
        scores = torch.matmul(user_embeds, item_embeds.t()) #[B,H]
        scores, indices = torch.topk(scores, k=k, dim=-1)
        return scores, indices

    def full_scores(self, user_dict, items):
        user_embeds = self.get_user_embedding(user_dict)  # [B, D]
        item_embeds = self.item_embedding(items)  # [H, D]
        scores = torch.matmul(user_embeds, item_embeds.t())  # [B,H]
        return scores

    def _init_weights(self, module):
        if isinstance(module, nn.Embedding):
            normal_(module.weight.data, mean=0.0, std=0.01)

    def gather_indexes(self, output, gather_index):
        """Gathers the vectors at the specific positions over a minibatch"""
        gather_index = gather_index.view(-1, 1, 1).expand(-1, -1, output.shape[-1])
        output_tensor = output.gather(dim=1, index=gather_index)
        return output_tensor.squeeze(1)

    def bpr_loss(self, pos_score, neg_score):
        return -torch.log(1e-10 + torch.sigmoid(pos_score - neg_score))


import numpy as np
from .Abstract_Ranker import Abstract_Reweigher

r"""
S-DRO
################################################

@inproceedings{10.1145/3485447.3512255,
author = {Wen, Hongyi and Yi, Xinyang and Yao, Tiansheng and Tang, Jiaxi and Hong, Lichan and Chi, Ed H.},
title = {Distributionally-robust Recommendations for Improving Worst-case User Experience},
year = {2022},
isbn = {9781450390965},
publisher = {Association for Computing Machinery},
address = {New York, NY, USA},
url = {https://doi.org/10.1145/3485447.3512255},
doi = {10.1145/3485447.3512255},
abstract = {Modern recommender systems have evolved rapidly along with deep learning models that are well-optimized for overall performance, especially those trained under Empirical Risk Minimization (ERM). However, a recommendation algorithm that focuses solely on the average performance may reinforce the exposure bias and exacerbate the “rich-get-richer” effect, leading to unfair user experience. In a simulation study, we demonstrate that such performance gap among various user groups is enlarged by an ERM-trained recommender in the long-term. To mitigate such amplification effects, we propose to optimize for the worst-case performance under the Distributionally Robust Optimization (DRO) framework, with the goal of improving long-term fairness for disadvantaged subgroups. In addition, we propose a simple-yet-effective streaming optimization improvement called Streaming-DRO (S-DRO), which effectively reduces loss variances for recommendation problems with sparse and long-tailed data distributions. Our results on two large-scale datasets suggest that (1) DRO is a flexible and effective technique for improving worst-case performance, and (2) Streaming-DRO outperforms vanilla DRO and other strong baselines by improving the worst-case and overall performance at the same time.},
booktitle = {Proceedings of the ACM Web Conference 2022},
pages = {3606–3610},
numpages = {5},
keywords = {Distributional robustness, Recommendation., Robust learning},
location = {Virtual Event, Lyon, France},
series = {WWW '22}
}


"""

class SDRO(Abstract_Reweigher):
    def __init__(self, config, group_weight):
        super().__init__(config)
        self.group_weight = group_weight

    def reset_parameters(self):
        self.mu = np.ones(self.config['group_num'])

    def reweight(self, input_dict):
        items = input_dict['items']
        losses = input_dict['loss'] #[B]

        adj_matrix = self.M[items]

        #C_t = np.mean(self.M[items], axis=1, keepdims=False) #[B,G]
        group_num = np.sum(adj_matrix, axis=0)
        gradient = np.sum(losses[:, np.newaxis] * adj_matrix, axis=0, keepdims=False) / (group_num + 1e-1)
        ###since the loss is the negative of the scores
        self.mu = self.mu + self.config['eta'] * gradient
        #self.mu = np.clip(self.mu, 1e-3, 1)
        self.mu = np.exp(self.mu)
        self.mu = self.mu / np.sum(self.mu)

        batch_weight = np.matmul(adj_matrix, self.mu * self.group_weight)
        batch_weight = batch_weight / np.sum(batch_weight)


        return batch_weight




from .IPS import IPS
from .SDRO import SDRO
from .Minmax_SGD import Minmax_SGD
from .APR import APR
from .FOCF import FOCF
from .FairDual import FairDual
from .Reg import Reg
from .FairNeg import FairNeg
from .DPR import DPR


from collections import *
import torch
import torch.nn.init as init
from torch import nn

from .base import BasePostProcessModel

MAXDOC = 50

'''
@inproceedings{qin2020diversifying,
  title={Diversifying search results using self-attention network},
  author={Qin, Xubo and Dou, Zhicheng and Wen, Ji-Rong},
  booktitle={Proceedings of the 29th ACM International Conference on Information \& Knowledge Management},
  pages={1265--1274},
  year={2020}
}
'''


class SelfAttnEnc(nn.Module):
    def __init__(self, d_model, nhead, nlayers, dropout):
        """
        Initialization
        
        :param d_model: The dimension of the model.
        :param nhead: The number of attention heads.
        :param nlayers: The number of transformer encoder layers.
        :param dropout: Dropout probability.
        """
        super(SelfAttnEnc, self).__init__()
        self.enc_layer = nn.TransformerEncoderLayer(d_model=d_model, nhead=nhead, dim_feedforward=400, dropout=dropout,
                                                    batch_first=True)
        self.enc = nn.TransformerEncoder(self.enc_layer, num_layers=nlayers)

    def forward(self, input, mask):
        """
        Forward pass of the Self-Attention Encoder.

        :param input: Input tensor of shape (batch_size, sequence_length, d_model).
        :param mask: Mask tensor indicating padded positions.
        :return: Encoded output tensor of shape (batch_size, sequence_length, d_model).
        """
        enc_out = self.enc(input, src_key_padding_mask=mask) # input [bs , seq_len , d_model]
        return enc_out # enc_out [bs , seq_len , d_model]


class DESA(BasePostProcessModel):
    def __init__(self, doc_d_model, doc_nhead, doc_nlayers, sub_d_model, sub_nhead, sub_nlayers, nhead, dropout):
        """
        Initialization

        :param doc_d_model: Document embedding dimension.
        :param doc_nhead: Number of attention heads for document encoder.
        :param doc_nlayers: Number of transformer layers for document encoder.
        :param sub_d_model: Sub-document embedding dimension.
        :param sub_nhead: Number of attention heads for sub-document encoder.
        :param sub_nlayers: Number of transformer layers for sub-document encoder.
        :param nhead: Number of attention heads for multi-head attention.
        :param dropout: Dropout probability.
        """
        super().__init__(dropout)
        
        Linear_out = 128
        self.linear1 = nn.Linear(doc_d_model, Linear_out)
        self.linear2 = nn.Linear(sub_d_model, Linear_out)
        self.linear3 = nn.Linear(18, 1)
        self.linear4 = nn.Linear(18+Linear_out+Linear_out+10, 1)
        self.doc_attn = SelfAttnEnc(Linear_out, doc_nhead, doc_nlayers, dropout) # [bs , seq_len , d_model]
        self.sub_attn = SelfAttnEnc(Linear_out, sub_nhead, sub_nlayers, dropout) # [bs , seq_len , d_model]
        self.dec_attn = nn.MultiheadAttention(Linear_out, nhead, dropout=dropout, batch_first=True)

    def fit(self, doc_emb, sub_emb, doc_mask, sub_mask, pos_qrel_feat, pos_subrel_feat, index_i=None, index_j=None,
                neg_qrel_feat=None, neg_subrel_feat=None, subrel_mask=None, mode='Train'):
        """
        Model training.

        :param doc_emb: Document embeddings of shape (batch_size, sequence_length, embedding_dim).
        :param sub_emb: Sub-document embeddings of shape (batch_size, sequence_length, embedding_dim).
        :param doc_mask: Mask tensor for document sequences.
        :param sub_mask: Mask tensor for sub-document sequences.
        :param pos_qrel_feat: Positive query relevance features.
        :param pos_subrel_feat: Positive sub-document relevance features.
        :param index_i: Index tensor for selecting positive samples.
        :param index_j: Index tensor for selecting negative samples.
        :param neg_qrel_feat: Negative query relevance features (optional).
        :param neg_subrel_feat: Negative sub-document relevance features (optional).
        :param subrel_mask: Mask tensor for sub-document relevance (optional).
        :param mode: Mode of operation ('Train' or 'Eval').
        :return: Positive and negative ranking scores in training mode, or final scores in evaluation mode.
        """
        doc_mask, sub_mask = doc_mask.bool(), sub_mask.bool()
        doc_rep = self.doc_attn(self.linear1(doc_emb), doc_mask)  # [bs, sq(50), d_model]
        sub_rep = self.sub_attn(self.linear2(sub_emb), sub_mask)  # [bs, sq(10), d_model]
        doc_dec, _ = self.dec_attn(doc_rep, sub_rep, sub_rep) # [bs, sq(50), d_model]
        device = doc_rep.device
        if mode == 'Train':
            pos_index_select1 = torch.index_select(doc_rep.reshape((-1, doc_rep.shape[2])), 0,
                                                   (index_i.to(device) + torch.linspace(0, doc_rep.shape[0] - 1,
                                                doc_rep.shape[0]).to(device) * torch.tensor(
                                                       doc_rep.shape[0] + 1).to(device)).long())
            pos_index_select2 = torch.index_select(doc_dec.reshape((-1, doc_dec.shape[2])), 0,
                                                   (index_i.to(device) + torch.linspace(0, doc_dec.shape[0] - 1,
                                                   doc_dec.shape[0]).to(device) * torch.tensor(
                                                       doc_dec.shape[0] + 1).to(device)).long())
            pos_concat = torch.cat([pos_qrel_feat, pos_index_select1, pos_index_select2,
                                    self.linear3(pos_subrel_feat).squeeze(2)], dim=1)  # pos_subrel[bs, sq(10), 18]
            pos_out = self.linear4(pos_concat)
            neg_index_select1 = torch.index_select(doc_rep.reshape((-1, doc_rep.shape[2])), 0,
                    (index_j.to(device) + torch.linspace(0, doc_rep.shape[0] - 1, doc_rep.shape[0]).to(device) *
                     torch.tensor(doc_rep.shape[0] + 1).to(device)).long())
            neg_index_select2 = torch.index_select(doc_dec.reshape((-1, doc_dec.shape[2])), 0,
                    (index_j.to(device) + torch.linspace(0, doc_dec.shape[0] - 1, doc_dec.shape[0]).to(device) * torch.tensor(
                                                                doc_dec.shape[0] + 1).to(device)).long())

            neg_concat = torch.cat([neg_qrel_feat, neg_index_select1,
                                    neg_index_select2, self.linear3(neg_subrel_feat).squeeze(2)], dim=1)
            neg_out = self.linear4(neg_concat)
            return pos_out, neg_out
        else:
            pos_concat = torch.cat([pos_qrel_feat, doc_rep, doc_dec,
                                    self.linear3(pos_subrel_feat).squeeze(3)], dim=2)  # pos_subrel[bs, sq(10), 18]
            pos_out = self.linear4(pos_concat)
            return pos_out.squeeze(2).squeeze(0) # [50]


class MLP(nn.Module):
    """
    Multi-Layer Perceptron (MLP).
    """
    def __init__(self, input_size, hid_size, output_size):
        """
        Initialization

        :param input_size: The size of the input feature vector.
        :param hid_size: The number of hidden units.
        :param output_size: The number of output units.
        """
        super(MLP, self).__init__()
        self.mlp = nn.Sequential(
            nn.Linear(input_size, hid_size),
            nn.ReLU(),
            nn.Linear(hid_size, output_size)
        )

    def forward(self, input):
        """
        Forward of the MLP.

        :param input: Input feature vector.
        :return: Output tensor.
        """
        output = self.mlp(input)
        return output



import os
import yaml
import json
import numpy as np
import torch
from ..utils import Build_Adjecent_Matrix



class Abstract_Reweigher(object):
    def __init__(self, config):
        self.config = config

        self.M, self.iid2pid = Build_Adjecent_Matrix(config)
        self.IR_type = ["ranking", "retrieval"]
        self.fair_type = "re-weight"
        self.type = ["point", "pair", "sequential"]

    def reset_parameters(self, **kwargs):
        self.exposure_count = np.zeros(self.config['group_num'])


    def reweight(self, **kwargs):
        """
            Recalculates the batch weights based on the loss and the group adjacency matrix.

            This function computes new weights for each group in the batch based on the loss associated with each item and
            the group adjacency matrix. The batch weights are recalculated by considering the loss for each group and adjusting
            them based on predefined parameters in the configuration (e.g., `beta`, `alpha`, `eta`).

            :param Any:
            :return: A normalized vector of batch weights for each group in the batch.
        """
        pass


class Abstract_Regularizer(object):
    def __init__(self, config):
        self.type = ["pair", "sequential"]
        self.IR_type = ["ranking", "retrieval"]
        self.fair_type = "regularizer"
        self.config = config
        self.M, self.iid2pid = Build_Adjecent_Matrix(config)

    def fairness_loss(self, **kwargs):
        """
           Computes the fairness loss as the variance of the input scores.

           This function calculates the fairness loss by computing the variance of the provided scores. The loss aims to capture
           the variation in the scores, which can be used as a measure of fairness in the model's predictions.

           :param: any.
           :return: The fairness loss.
           """
        pass

class Abstract_Sampler(object):
    def __init__(self, config, user2pos):
        self.user2pos = user2pos ###record the negative item corpus
        self.type = ["pair"]
        self.IR_type = ["ranking", "retrieval"]
        self.fair_type = "sample"
        self.config = config
        self.M, self.iid2pid = Build_Adjecent_Matrix(config)
        self.group_sizes = np.sum(self.M, axis=0)
        #print(self.group_sizes)
        #exit(0)

    def reset_parameters(self, **kwargs):
        self.sample_probality = np.ones(self.config['group_num'])
        self.sample_probality = self.sample_probality/np.sum(self.sample_probality)

    def sample(self, **kwargs):
        """
        Samples negative items for each user in the interaction, based on group fairness and model predictions.

        This function selects negative items for users in the interaction based on two factors:
        group fairness (weighted by `group_fair`) and the model's predicted scores (scaled by a temperature parameter).
        The negative item selection combines these factors and returns the selected items.

        :param interaction: A dictionary containing interaction data. It should include:
                            - 'user_ids': A tensor of user IDs.
                            - 'item_ids': A tensor of item IDs.
        :param Model: The model object that provides the method `full_scores` to get predicted scores for negative items.
        :return: A tensor of sampled negative item IDs for each user, and the positive item embeddings.
        """


        pass



import numpy as np
from .Abstract_Ranker import Abstract_Regularizer
import torch.nn.functional as F
import torch
r"""
FOCF
################################################


@article{yao2017beyond,
  title={Beyond parity: Fairness objectives for collaborative filtering},
  author={Yao, Sirui and Huang, Bert},
  journal={Advances in neural information processing systems},
  volume={30},
  year={2017}
}

##Note that in this code, we extend the two-group cases into multi-group cases

"""

class FOCF(Abstract_Regularizer):
    def __init__(self, config, group_weight):
        super().__init__(config)


    def fairness_loss(self, input_dict):
        losses = input_dict['scores'] #[B]
        mean_loss = torch.mean(losses)
        gap = torch.abs(losses-mean_loss)
        return F.smooth_l1_loss(gap, torch.zeros_like(gap))






import os
import math
import json
import pickle
import numpy as np
from collections import Counter
from tqdm import tqdm

from .utils import pkl_save

MAXDOC = 50


def generate_bm25_scores_for_query(config):
    """
    Generates BM25 relevance scores for queries and their suggested variations against documents.
    This function processes each query and its suggested alternatives, calculating BM25 scores 
    against a collection of documents. The scores are computed for both the original query and 
    its suggestions, then saved to a pickle file.

    :param config: A dictionary containing configuration parameters including data directories
                  and model settings.
    """
    qd = pickle.load(open(os.path.join(config['data_dir'], 'div_query.data'), 'rb'))
    
    data_content_dir = config['data_content_dir']
    doc_content_dict = {}
    for fname in os.listdir(data_content_dir):
        with open(os.path.join(data_content_dir, fname), 'r') as fr:
            data_dict = json.load(fr)
            doc_content_dict[fname[:-5]] = data_dict

    bm25_dict = {}
    for qid in tqdm(qd):
        bm25_dict[qid] = [{}]
        query = qd[qid].query
        doc_list = qd[qid].doc_list[:MAXDOC]
        query_suggestion = qd[qid].query_suggestion
        doc_content_list = [doc_content_dict[doc_id]['body_text'] for doc_id in doc_list]
        bm25_scores = calculate_bm25(query, doc_content_list)
        for i, doc_id in enumerate(doc_list):
            bm25_dict[qid][0][doc_id] = bm25_scores[i]
        sub_query_bm25 = []
        for sub_query in query_suggestion:
            bm25_scores = calculate_bm25(sub_query, doc_content_list)
            sub_query_bm25.append(bm25_scores)
        bm25_dict[qid].append(np.array(sub_query_bm25))
    
    if not os.path.exists(os.path.join(config['data_dir'], config['model'])):
        os.makedirs(os.path.join(config['data_dir'], config['model']))
    pkl_save(bm25_dict, os.path.join(config['data_dir'], config['model'], 'bm25_scores.pkl'))


def calculate_bm25(query, documents, k1=1.5, b=0.75):
    """
    Calculates BM25 relevance scores between a query and a list of documents.
    This function implements the BM25 ranking algorithm, which combines term frequency,
    inverse document frequency, and document length normalization to score document
    relevance to a query.

    :param query: A string containing the search query
    :param documents: A list of strings, where each string is a document's text
    :param k1: Float parameter controlling term frequency scaling (default: 1.5)
    :param b: Float parameter controlling document length normalization (default: 0.75)
    :return: A list of float values representing BM25 scores for each document
    """
    N = len(documents)
    avgdl = sum(len(doc.split()) for doc in documents) / N

    bm25_scores = []

    def idf(term):
        df = sum(1 for doc in documents if term in doc.split())
        return math.log((N - df + 0.5) / (df + 0.5) + 1.0)

    for doc in documents:
        score = 0.0
        doc_terms = Counter(doc.split())
        doc_length = len(doc.split())

        for term in query.split():
            if term in doc_terms:
                f = doc_terms[term]
                term_idf = idf(term)
                score += term_idf * (f * (k1 + 1)) / (f + k1 * (1 - b + b * (doc_length / avgdl)))

        bm25_scores.append(score)
    
    return bm25_scores



import pandas as pd
import os

"""
@article{yang2020causal,
  title={Causal intersectionality for fair ranking},
  author={Yang, Ke and Loftus, Joshua R and Stoyanovich, Julia},
  journal={arXiv preprint arXiv:2006.08688},
  year={2020}
}

License
Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated 
documentation files (the "Software"), to deal in the Software without restriction, including without limitation 
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, 
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE 
OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

def get_counterfactual_data_real(data, model_path, configs, dataset):
    """
    Append to data the fair columns containing the counterfactual values
    computed based on the causal estimates saved at model_path

    Args:
        data (pandas.DataFrame): The current DataFrame containing the data to be transformed.
        model_path (str): Path to the output file where the model and predictions will be saved
        configs (dict): Configuration dict of the fairness method
        data_configs (dict): Configuration dict of the dataset

    Returns:
        pandas.DataFrame: The adjusted DataFrame with counterfactual columns added.
    """
    group_list = [x for x in data[configs['group']].unique() if x != configs['control']]

    mediators = pd.read_csv(os.path.join(model_path, "identified_mediators.csv"))
    no_mediators = len(mediators) == 0 or str(mediators['mediators'].values[0]) == 'nan'

    new_cols = []
    for med in dataset.feature_cols:
        if med in mediators['mediators'].values:
            x_res = pd.read_csv(os.path.join(model_path, med + "~" + configs['group'] + "-1.csv"))
            counter_g_base = x_res[x_res["Unnamed: 0"] == configs['group'] + configs['control']]["Estimate"].values[0]

            x_shifts = {configs['control']: 0}
            for gi in group_list:
                if not configs['group'] + gi in x_res["Unnamed: 0"].values:
                    x_shifts[gi] = 0
                else:
                    other_g_base = x_res[x_res["Unnamed: 0"] == configs['group'] + gi]["Estimate"].values[0]
                    x_shifts[gi] = counter_g_base - other_g_base

            feature_shifts = data[configs['group']].apply(lambda x: x_shifts[x])
            data.loc[:, med + "_fair"] = data[med] + feature_shifts
            new_cols.append(med + "_fair")

        else:
            # variables that are not mediators remain unchanged
            data.loc[:, med + "_fair"] = data[med]
            new_cols.append(med + "_fair")

    if no_mediators:
        # direct effect of the IV on the DV --> we keep the observed X as it is
        y_res = pd.read_csv(os.path.join(model_path, dataset.score_col + '~' + configs['group'] + "-1.csv"))
        counter_g_base = y_res[y_res["Unnamed: 0"] == configs['group'] + configs['control']]["Estimate"].values[0]
        y_shifts = {configs['control']: 0}
        for gi in group_list:
            if not configs['group'] + gi in y_res["Unnamed: 0"].values:
                y_shifts[gi] = 0
            else:
                y_shifts[gi] = counter_g_base - y_res[y_res["Unnamed: 0"] == configs['group'] + gi]["Estimate"].values[0]
    else:
        y_shifts = {configs['control']: 0}
        y_shifts_resolve = {configs['control']: 0}
        for gi in group_list:
            if not os.path.exists(os.path.join(model_path, gi + "_med" + ".csv")):
                y_shifts[gi] = 0
                y_shifts_resolve[gi] = 0
            else:
                g_res = pd.read_csv(os.path.join(model_path, gi + "_med" + ".csv"))
                y_shifts[gi] = -g_res[g_res['Metric'] == 'Total Effect']["Estimate"].values[0]
                y_shifts_resolve[gi] = -g_res[g_res['Metric'] == 'Direct Effect']["Estimate"].values[0]

    data["Y_shift"] = data[configs['group']].apply(lambda x: y_shifts[x])
    data[dataset.score_col + "_fair"] = data[dataset.score_col] + data["Y_shift"]
    new_cols.append(dataset.score_col + "_fair")

    return data


import numpy as np
from .Abstract_Reranker import Abstract_Reranker
from gurobipy import Model, GRB, quicksum

r"""
RAIF, a model-agnostic repeat-bias-aware item fairness optimization (RAIF) algorithm based on MILP.
Note that we remove repeat bias term, change the item fairness objective to make the exposure of each group closer, and extend RAIF into multi-group cases.

@article{liu2025repeat,
  title={Repeat-bias-aware Optimization of Beyond-accuracy Metrics for Next Basket Recommendation},
  author={Liu, Yuanna and Li, Ming and Aliannejadi, Mohammad and de Rijke, Maarten},
  journal={arXiv e-prints},
  pages={arXiv--2501},
  year={2025}
}

"""


def get_results(num_users, size, topk, solution, topk_items):
    """
    Converts the solution matrix into selected item lists for multiple users.

    Parameters:
    ----------
    num_users: int
        The number of users.
    size: int
        The expected number of items per user in the final rerank list.
    topk: int
        The number of candidate items per user.
    solution: numpy.ndarray, shape (num_users, topk)
        A matrix indicating the final selected items.
    topk_items: list of list of int, shape (num_users, topk)
        A list where each entry contains candidate item IDs corresponding to a user.

    Returns:
    -------
    rerank: list of list of int, shape (num_users, size)
        A list where each entry contains exactly `size` selected items for a user.
    """
  
    rerank = []
    for i in range(num_users):

        rerank_user = []
        for j in range(topk):
            if solution[i, j] > 0.5:
                rerank_user.append(topk_items[i][j])

        assert len(rerank_user) == size
        rerank.append([int(x) for x in rerank_user])

    return rerank

def load_ranking_matrices(relevance, topk): 
    """
    Generates ranking matrices by selecting the top-k relevant items for each user.

    Parameters:
    ----------
    relevance: numpy.ndarray, shape (num_users, num_items)
        A 2D array where each row corresponds to a user and contains item relevance scores.
    topk: int
        The number of top-ranked items to select per user.

    Returns:
    -------
    topk_items: numpy.ndarray, shape (num_users, topk)
        A 2D array where each row contains the indices of the top-k items for the corresponding user.
    topk_scores: numpy.ndarray, shape (num_users, topk)
        A 2D array where each row contains the relevance scores of the selected top-k items.
    num_users: int
        The total number of users.

    """
  
    num_users, num_items = relevance.shape
    
    topk_items = np.zeros((num_users, topk), dtype=int)
    topk_scores = np.zeros((num_users, topk))

    for user_idx in range(num_users):
        # Get the indices of the items sorted by their relevance score in descending order
        sorted_indices = np.argsort(relevance[user_idx])[::-1]
        
        # Select the top k indices and corresponding scores
        topk_items[user_idx] = sorted_indices[:topk]
        topk_scores[user_idx] = relevance[user_idx, sorted_indices[:topk]]
    
    return topk_items, topk_scores, num_users


def read_item_index(total_users, topk, no_item_groups, item_group_map, topk_items):
    """
    Creates a binary indicator matrix that maps items to their respective item groups.

    Parameters:
    ----------
    total_users: int
        The total number of users.
    topk: int
        The number of candidate items per user.
    no_item_groups: int
        The total number of item groups.
    item_group_map: dict
        A dictionary mapping item indices to their corresponding group IDs.
    topk_items: list of list of int, shape (total_users, topk)
        A list where each entry contains candidate item IDs corresponding to a user.

    Returns:
    -------
    Ihelp: numpy.ndarray, shape (total_users, topk, no_item_groups)
        A binary 3D array where `Ihelp[uid][lid][k] = 1` if the `lid`-th item for user `uid`
        belongs to item group `k`, otherwise `0`.
    """
    Ihelp = np.zeros((total_users, topk, no_item_groups))
    for uid in range(total_users):
        for lid in range(topk):
            for k in range(no_item_groups):
                if item_group_map[topk_items[uid][lid]] == k:
                    Ihelp[uid][lid][k] = 1

    return Ihelp


def fairness_optimisation(total_users, alpha, size, topk, group_num, Ihelp, topk_scores, mean):
    """
    Solves a fairness-aware ranking optimization problem using Gurobi.

    Parameters:
    ----------
    total_users: int
        The total number of users.
    alpha: float
        The fairness regularization parameter. A higher alpha increases fairness consideration.
    size: int
        The number of items to be selected per user.
    topk: int
        The number of candidate items per user.
    group_num: int
        The number of item groups.
    Ihelp: numpy.ndarray, shape (total_users, topk, group_num)
        A binary indicator matrix.
    topk_scores: numpy.ndarray, shape (total_users, topk)
        A 2D relevance score matrix.
    mean: float
        The mean exposure across item groups.

    Returns:
    -------
    solution: numpy.ndarray, shape (num_users, topk)
        A matrix indicating the final selected items.
  
    """
  

    print(f"Running RAIF, {format(alpha, 'f')}")
    # V1: No. of users
    # V2: No. of top items (topk)
    # V4: no. of item groups
    V1, V2, V4 = range(total_users), range(topk), range(group_num)

    # initiate model
    model = Model()

    W = model.addVars(V1, V2, vtype=GRB.BINARY)
    item_group = model.addVars(V4, vtype=GRB.CONTINUOUS)
    item_fair = model.addVar(vtype=GRB.CONTINUOUS)
    abs_diff = model.addVars(V4, lb=0, name="abs_diff")
               
    model.setObjective(quicksum(topk_scores[i][j] * W[i, j] for i in V1 for j in V2) - alpha * item_fair, GRB.MAXIMIZE)

    for i in V1:
        model.addConstr(quicksum(W[i, j] for j in V2) == size)
    
    for k in V4:
        model.addConstr(item_group[k] == quicksum(W[i, j] * Ihelp[i][j][k] for i in V1 for j in V2))
    
    for k in V4:
        model.addConstr(abs_diff[k] >= item_group[k] - mean)
        model.addConstr(abs_diff[k] >= -(item_group[k] - mean))

    model.addConstr(item_fair == quicksum(abs_diff[k] for k in V4))


    # optimizing
    model.optimize()
    if model.status == GRB.OPTIMAL:
        solution = model.getAttr('x', W)
        #fairness = model.getAttr('x', item_group)


    return solution



class RAIF(Abstract_Reranker):
    def __init__(self, config, weights = None):
        super().__init__(config, weights)


    def rerank(self, ranking_score, k):
        ## its parameters
        topk = self.config['candidate']
        alpha = self.config['alpha']

        topk_items, topk_scores, num_users = load_ranking_matrices(ranking_score, topk)
        mean = (num_users * k) / self.group_num
        Ihelp = read_item_index(total_users=num_users, topk=topk, no_item_groups=self.group_num, item_group_map=self.iid2pid, topk_items=topk_items) 
        solution = fairness_optimisation(num_users, alpha, k, topk, self.group_num, Ihelp, topk_scores, mean)
        rerank_list = get_results(num_users, k, topk, solution, topk_items)
        
        return rerank_list




import numpy as np
from .Abstract_Ranker import Abstract_Reweigher

r"""
Min-max-SGD
################################################


@InProceedings{pmlr-v162-abernethy22a,
  title = 	 {Active Sampling for Min-Max Fairness},
  author =       {Abernethy, Jacob D and Awasthi, Pranjal and Kleindessner, Matth{\"a}us and Morgenstern, Jamie and Russell, Chris and Zhang, Jie},
  booktitle = 	 {Proceedings of the 39th International Conference on Machine Learning},
  pages = 	 {53--65},
  year = 	 {2022},
  editor = 	 {Chaudhuri, Kamalika and Jegelka, Stefanie and Song, Le and Szepesvari, Csaba and Niu, Gang and Sabato, Sivan},
  volume = 	 {162},
  series = 	 {Proceedings of Machine Learning Research},
  month = 	 {17--23 Jul},
  publisher =    {PMLR},
  pdf = 	 {https://proceedings.mlr.press/v162/abernethy22a/abernethy22a.pdf},
  url = 	 {https://proceedings.mlr.press/v162/abernethy22a.html},
  abstract = 	 {We propose simple active sampling and reweighting strategies for optimizing min-max fairness that can be applied to any classification or regression model learned via loss minimization. The key intuition behind our approach is to use at each timestep a datapoint from the group that is worst off under the current model for updating the model. The ease of implementation and the generality of our robust formulation make it an attractive option for improving model performance on disadvantaged groups. For convex learning problems, such as linear or logistic regression, we provide a fine-grained analysis, proving the rate of convergence to a min-max fair solution.}
}


"""

class Minmax_SGD(Abstract_Reweigher):
    def __init__(self, config, group_weight):
        super().__init__(config)
        self.group_weight = group_weight

    def reset_parameters(self):
        pass

    def reweight(self, input_dict):
        items = input_dict['items']
        losses = input_dict['loss'] #[B]

        adj_matrix = self.M[items]
        #loss_groups = np.mean(losses[:, np.newaxis] * adj_matrix, axis=0, keepdims=False)
        group_num = np.sum(adj_matrix, axis=0)
        loss_groups = np.sum(losses[:, np.newaxis] * adj_matrix, axis=0, keepdims=False) / (group_num + 1e-1)

        worst_groups = np.argmax(loss_groups)
        weights = np.zeros(self.config['group_num']) + self.config['p']
        weights[worst_groups] = 1.0 * (1-self.config['p'])
        batch_weight = np.matmul(adj_matrix, weights * self.group_weight)
        #batch_weight = batch_weight / np.sum(batch_weight)

        return batch_weight




import os
import time
import pickle
import torch
import numpy as np
from tqdm import tqdm
from torch.utils.data import Dataset, DataLoader
from torch.nn.utils import clip_grad_norm_
from ..postprocessing_model.DESA import DESA
from ..utils.loss import list_pairwise_loss
from ..post_evaluator import evaluate_test_qids_DESA


class DESADataset(Dataset):
    """
    A PyTorch Dataset class for handling training data, ensuring they are properly prepared for model training.

    :param input_list: List of training samples containing input tensors and features
    """
    def __init__(self,  input_list):
        self.data = input_list

    def __len__(self):
        return len(self.data)

    def __getitem__(self,  idx):
        doc_emb = self.data[idx][0].float()
        sub_emb = self.data[idx][1].float()
        doc_mask = 1-self.data[idx][2]
        sub_mask = 1-self.data[idx][3]
        weight = torch.tensor(self.data[idx][4]).float()
        index_i = self.data[idx][5]
        index_j = self.data[idx][6]
        pos_qrel_feat = self.data[idx][7].float()
        neg_qrel_feat = self.data[idx][8].float()
        pos_subrel_feat = self.data[idx][9].float()
        neg_subrel_feat = self.data[idx][10].float()
        subrel_mask = 1-self.data[idx][11]
        doc_emb.requires_grad = False
        sub_emb.requires_grad = False
        doc_mask.requires_grad = False
        sub_mask.requires_grad = False
        weight.requires_grad = False
        index_i.requires_grad = False
        index_j.requires_grad = False
        pos_qrel_feat.requires_grad = False
        neg_qrel_feat.requires_grad = False
        pos_subrel_feat.requires_grad = False
        neg_subrel_feat.requires_grad = False
        subrel_mask.requires_grad = False
        return doc_emb, sub_emb, doc_mask, sub_mask, weight, index_i, index_j, pos_qrel_feat, neg_qrel_feat, \
               pos_subrel_feat, neg_subrel_feat, subrel_mask


def get_fold_loader(fold, train_data, BATCH_SIZE):
    """
    DataLoader for a specific cross-validation fold's training data.
    
    :param fold: Current fold number in cross-validation
    :param train_data: Training data dictionary containing document and suggestion features
    :param BATCH_SIZE: Number of samples per batch
    :return: DataLoader object containing the processed training data
    """
    input_list = []
    starttime = time.time()
    print('Begin loading fold {} training data'.format(fold))
    for i in range(len(train_data)):
        input_list.append([train_data[i]['doc2vec'],
                           train_data[i]['sub2vec'],
                           train_data[i]['doc2vec_mask'],
                           train_data[i]['sub2vec_mask'],
                           train_data[i]['weight'],
                           torch.argmax(train_data[i]['positive_mask']),
                           torch.argmax(train_data[i]['negative_mask']),
                           train_data[i]['pos_qrel_feat'],
                           train_data[i]['neg_qrel_feat'],
                           train_data[i]['pos_subrel_feat'],
                           train_data[i]['neg_subrel_feat'],
                           train_data[i]['subrel_feat_mask']])
    desa_dataset = DESADataset(input_list)
    loader = DataLoader(
        dataset=desa_dataset,
        batch_size=BATCH_SIZE,
        shuffle=True,
        num_workers=4,
        pin_memory=True,
    )
    print('Training data loaded!')
    endtime = time.time()
    print('Total time  = ', round(endtime - starttime, 2), 'secs')
    return loader


def DESA_run(config):
    """
    Executes the complete training and evaluation pipeline for the DESA model.
    
    :param config: Dictionary containing model configuration parameters
    """

    # load randomly shuffled queries
    if not os.path.exists(os.path.join(config['model_save_dir'], config['model'])):
        os.makedirs(os.path.join(config['model_save_dir'], config['model']))
    qd = pickle.load(open(os.path.join(config['data_dir'], 'div_query.data'), 'rb'))
    fold_p = os.path.join(config['data_dir'], config['model'], 'fold/')
    final_metrics = []
    best_model_list = []
    fold_time = 0
    for _fold in os.listdir(fold_p):
        train_data = torch.load(os.path.join(fold_p, _fold, 'train_data.pkl'))
        test_data = torch.load(os.path.join(fold_p, _fold, 'test_data.pkl'))
        desa_data_loader = get_fold_loader(fold_time, train_data, config['batch_size'])
        fold_time += 1
        print('Fold = ', fold_time)
        model = DESA(config['embedding_length'], 8, 2,config['embedding_length'], 8, 2, 8, config['dropout'])
        # doc_d_model, doc_nhead, doc_nlayers, sub_d_model, sub_nhead, sub_nlayers, d_model, nhead
        if torch.cuda.is_available():
            model = model.cuda()
        opt = torch.optim.Adam(model.parameters(), lr=config['learning_rate'], betas=(0.9, 0.999), eps=1e-08, weight_decay=5e-4)
        params = list(model.parameters())
        if fold_time == 1:
            print('model = ', model)
            print(len(params))
            for param in params:
                print(param.size())
            n_params = sum([p.numel() for p in model.parameters() if p.requires_grad])
            print('* number of parameters: %d' % n_params)

        max_metric = 0
        patience = 0
        best_model = ""
        for epoch in range(config['epoch']):
            print('Start Training...')
            model.train()
            for step, train_data in enumerate(tqdm(desa_data_loader, desc='BATCH', ncols=80)):
                doc_emb, sub_emb, doc_mask, sub_mask, weight, index_i, index_j, \
                pos_qrel_feat, neg_qrel_feat, \
                pos_subrel_feat, neg_subrel_feat, subrel_mask = train_data
                if torch.cuda.is_available():
                    doc_emb = doc_emb.cuda()
                    sub_emb = sub_emb.cuda()
                    doc_mask = doc_mask.cuda()
                    sub_mask = sub_mask.cuda()
                    weight = weight.cuda()
                    index_i = index_i.cuda()
                    index_j = index_j.cuda()
                    pos_qrel_feat = pos_qrel_feat.cuda()
                    neg_qrel_feat = neg_qrel_feat.cuda()
                    pos_subrel_feat = pos_subrel_feat.cuda()
                    neg_subrel_feat = neg_subrel_feat.cuda()
                    subrel_mask = subrel_mask.cuda()
                score_1, score_2 = model.fit(doc_emb, sub_emb, doc_mask, sub_mask, pos_qrel_feat, pos_subrel_feat,
                                         index_i, index_j, neg_qrel_feat, neg_subrel_feat, subrel_mask)
                acc, loss = list_pairwise_loss(score_1, score_2, weight)
                opt.zero_grad()
                loss.backward()
                clip_grad_norm_(model.parameters(), max_norm=1)
                opt.step()
                if (step + 1) % config['eval_steps'] == 0:
                    model.eval()
                    metrics = []
                    for qid in test_data:
                        metric = evaluate_test_qids_DESA(model, test_data[str(qid)], qd[str(qid)])
                        metrics.append(metric)
                    avg_alpha_NDCG = np.mean(metrics)
                    if max_metric < avg_alpha_NDCG:
                        max_metric = avg_alpha_NDCG
                        tqdm.write('max avg_alpha_NDCG updated: {}'.format(max_metric))
                        model_filename = os.path.join(config['model_save_dir'], config['model'], 'TOTAL_EPOCH_' + str(config['epoch']) + '_FOLD_' + str(fold_time) + '_EPOCH_' + str(epoch) + '_LR_' + str(config['learning_rate']) + '_BATCHSIZE_' + str(
                        config['batch_size']) + '_DROPOUT_' + str(config['dropout']) + '_' + str(config['embedding_type']) + '.pickle')
                        torch.save(model.state_dict(), model_filename)
                        best_model = model_filename
                        patience = 0
                    else:
                        patience += 1
                    model.train()
                    if epoch > 0 and patience > 2:
                        new_lr = 0.0
                        for param_group in opt.param_groups:
                            param_group['lr'] = param_group['lr'] * 0.5
                            new_lr = param_group['lr']
                        patience = 0
                        tqdm.write("decay lr: {}, load model: {}".format(new_lr, best_model))
            model.eval()
            metrics = []
            for qid in test_data:
                metric = evaluate_test_qids_DESA(model, test_data[str(qid)], qd[str(qid)])
                metrics.append(metric)
            avg_alpha_NDCG = np.mean(metrics)
            if max_metric < avg_alpha_NDCG:
                max_metric = avg_alpha_NDCG
                tqdm.write('max avg_alpha_NDCG updated: {}'.format(max_metric))
                model_filename = os.path.join(config['model_save_dir'], config['model'], 'TOTAL_EPOCH_' + str(config['epoch']) + '_FOLD_' + str(fold_time) + '_EPOCH_' + str(epoch) + '_LR_' + str(config['learning_rate']) + '_BATCHSIZE_' + str(
                config['batch_size']) + '_DROPOUT_' + str(config['dropout']) + '_' + str(config['embedding_type']) + '.pickle')
                torch.save(model.state_dict(), model_filename)
                best_model = model_filename
            if epoch == (config['epoch'] - 1):
                final_metrics.append(max_metric)
                best_model_list.append(best_model)

    print('alpha-nDCG = {}, best model = {}'.format(sum(final_metrics)/len(final_metrics), best_model_list))





from .CIFRank import CIFRank
from .gFair import gFair
from .iFair import iFair
from .LFR import LFR

fairness_method_mapping = {
    'CIFRank': CIFRank,
    'gFair': gFair,
    'iFair': iFair,
    'LFR': LFR
}


import os
import numpy as np
from .Abstract_Reranker import Abstract_Reranker
from tqdm import tqdm,trange
import copy
r"""
Welf

Note that since the author do not publish the codes, we implement according to the equations of the paper, could be slightly different with the results
Welf is a item-side fairness algorithm, it may not perform good in multi-group fairness settings.

@article{do2021two,
  title={Two-sided fairness in rankings via Lorenz dominance},
  author={Do, Virginie and Corbett-Davies, Sam and Atif, Jamal and Usunier, Nicolas},
  journal={Advances in Neural Information Processing Systems},
  volume={34},
  pages={8596--8608},
  year={2021}
}
"""

def phi_func(x,alpha):
    if x >= 0:
        return np.power(x+1e-8,alpha)
    else:
        raise ValueError("x:",x)

class Welf(Abstract_Reranker):
    def __init__(self, config, weights = None):
        super().__init__(config, weights)


    def rerank(self, ranking_score, k):
        ## its parameters
        lambd = self.config['lambda']
        alpha1 = self.config['alpha1']
        alpha2 = self.config['alpha2']
        iter_num = self.config['iter_num']

        user_size = len(ranking_score)
        P_t = copy.copy(ranking_score)
        for iter in trange(iter_num):
            gamma = 2/(iter + 2)
            title_P = np.zeros((user_size, self.config['item_num']))
            B_l = np.zeros(self.config['group_num'])

            for t in range(user_size):
                #title_P = np.zeros((item_num,K))

                recommended_eatch_iter = np.argsort(P_t[t,:],axis=-1)[::-1]
                recommended_eatch_iter = recommended_eatch_iter[:k]

                B_l = B_l + np.sum(self.M[recommended_eatch_iter,:],axis=0,keepdims=False)
                title_P[t,:] = alpha1 * ranking_score[t,:] * phi_func(np.sum(ranking_score[t,:]), alpha1-1)
            #print(B_l.astype(np.int))
            for i in range(self.item_num):
                title_P[:,i] += lambd * alpha2 * np.ones(user_size) * phi_func(np.sum(self.M[i,:]*(1-B_l/(np.sum(B_l)+1e-5))), alpha2-1)

            P_t = (1-gamma) * P_t + gamma * title_P

        rerank_list = []
        for t in range(user_size):
            items = np.argsort(P_t[t,:],axis=-1)[::-1]
            x_allocation = items[:k]
            re_allocation = np.argsort(ranking_score[t,x_allocation])[::-1]
            x_allocation = x_allocation[re_allocation]
            rerank_list.append(x_allocation)

        return rerank_list



import os
from abc import ABC, abstractmethod
from pathlib import Path
from ..fairness_evaluator import evaluate, evaluate_runs


class Ranker(ABC):
    """
    Ranker class for training and evaluating ranking models.

    This abstract class provides methods for training ranking models and evaluating their fairness
    using the provided dataset and configuration settings.

    :param configs : dict
        Configuration dictionary containing paths and settings for the model and evaluation.
    :param dataset : object
        The dataset object containing data and metadata for training and evaluation.
    """

    def __init__(self, configs, dataset):
        """
        Initializes the Ranker class with the given configurations and dataset.

        :param configs : dict
            Configuration dictionary containing paths and settings for the model and evaluation.
        :param dataset : object
            The dataset object containing data and metadata for training and evaluation.
        """
        self.configs = configs
        self.dataset = dataset
        self.out_dir = Path(self.dataset.path) / self.configs["preprocessing_model"] / "ranker"
        self.ranker_path = Path(self.configs["ranker_path"])

    @abstractmethod
    def train(self, data_train, data_test):
        """
        Train the ranking model.

        This method is abstract and needs to be implemented in the subclasses to train a ranking model
        based on the provided training and testing data. The trained model is saved at a specific location.

        :param data_train : pandas.DataFrame
            The data to be used for training the ranking model.
        :param data_test : pandas.DataFrame
            The data to be used for testing the ranking model.

        :return : None
            This method does not return anything. It saves the trained model to a file.
        """
        pass

    @abstractmethod
    def predict(self, data, run, file_name):
        """
        Predict ranking scores for the given data.

        This method is abstract and should be implemented in subclasses to predict ranking scores
        based on the trained model. It appends the predicted scores to the provided data.

        :param data : pandas.DataFrame
            The data to be used for prediction.
        :param run : str
            The identifier for the specific run.
        :param file_name : str
            The name of the file where predictions will be saved.

        :return : pandas.DataFrame
            The data with an added column containing predicted ranking scores.
        """
        pass

    def evaluate_run(self, pred, run, file_name):
        """
        Evaluate the results of a specific run and compute fairness metrics.

        This method evaluates the predictions using fairness metrics for each sensitive attribute
        in the dataset and saves the evaluation results.

        :param pred : pandas.DataFrame
            The predictions to be evaluated.
        :param run : str
            The identifier for the specific run.
        :param file_name : str
            The name of the file containing predictions to evaluate.

        :return : None
            This method does not return anything. It saves the evaluation results to the file.
        """
        eval_path = Path(self.dataset.path) / self.configs["preprocessing_model"] / "evaluate" / str(run) / file_name
        print(eval_path)
        for s_attr in self.dataset.sensitive_cols:
            evaluate(pred, self.dataset.query_col, s_attr, eval_path, self.configs["evaluate"])

    def evaluate(self, file_name):
        """
        Evaluate multiple runs of the ranking model using fairness metrics.

        This method evaluates the model for all runs using the specified fairness metrics from the configuration.

        :param file_name : str
            The name of the file to evaluate.

        :return : None
            This method does not return anything. It performs the evaluation and saves the results.
        """
        runs = self.dataset.k_fold
        eval_path = Path(self.dataset.path) / self.configs["preprocessing_model"] / "evaluate"

        eval_files = [file for file in os.listdir(eval_path / "0" / file_name / self.configs["evaluate"]["metrics"][0])
                      if "QID" in file]
        qids = range(0, len(eval_files))
        evaluate_runs(self.configs["evaluate"], qids, eval_path, runs, file_name)



import numpy as np
from .Abstract_Ranker import Abstract_Reweigher
import cvxpy as cp
import time

r"""
FairDual
################################################
"""




class FairDual(Abstract_Reweigher):
    def __init__(self, config, group_weight):
        super().__init__(config)
        self.group_weight = group_weight

    def reset_parameters(self, train_len):
        self.train_len = train_len
        self.update_len = self.config['update_epoch'] * train_len
        self.C_t = self.config['s_k'] * self.update_len * self.group_weight
        self.mu = np.ones(self.config['group_num'])
        self.gradient_cusum = 0

        self.exposure_count = 0


    def reweight(self, input_dict):
        ###here the items are sampled items and predict the topk
        items = input_dict['sample_items']
        batch_size = len(items)

        B_t = np.sum(self.M[items], axis=1, keepdims=False)
        batch_weight = np.mean(B_t * self.mu, axis=-1, keepdims=False)
        batch_weight = 1.0 - batch_weight / np.sum(batch_weight)

        #############update parameters
        D_t = np.sum(B_t, axis=0, keepdims=False)
        self.C_t = self.C_t - D_t
        gradient = -D_t / (batch_size * self.config['s_k']) + self.C_t / (
                    self.train_len * self.config['s_k'])
        gradient = self.config['alpha'] * gradient + (
                    1 - self.config['alpha']) * self.gradient_cusum
        self.gradient_cusum = gradient

        for g in range(1):

            self.mu = self.compute_next_dual(self.config['eta'], self.group_weight, self.mu, gradient,
                                        self.config['lambd'])

        Dual_weight = batch_weight
        #print(self.mu)
        #batch_weight = np.exp(batch_weight) / np.sum(np.exp(batch_weight))

        # items = input_dict['items']
        #
        # adj_matrix = self.M[items]
        #
        # B_t = np.sum(adj_matrix, axis=0, keepdims=False)
        # self.exposure_count = self.exposure_count + B_t
        # norm_count = self.group_weight * self.exposure_count / np.sum(self.exposure_count)
        # batch_weight = np.matmul(adj_matrix, norm_count)
        # batch_weight = batch_weight / np.sum(batch_weight)
        # IPS_weight = 1/(batch_weight+0.1)


        return Dual_weight



    def compute_next_dual(self, eta, rho, dual, gradient, lambd):
        tilde_dual = dual - eta * gradient / rho / rho
        order = np.argsort(tilde_dual * rho)
        ordered_tilde_dual = tilde_dual[order]
        ordered_next_dual = self.map_layer(ordered_tilde_dual, rho[order], lambd)
        return ordered_next_dual[order.argsort()]

    def map_layer(self, ordered_tilde_dual, rho, lambd):
        m = len(rho)
        answer = cp.Variable(m)
        objective = cp.Minimize(cp.sum_squares(cp.multiply(rho, answer) - cp.multiply(rho, ordered_tilde_dual)))
        # objective = cp.Minimize(cp.sum(cp.multiply(rho,answer) - cp.multiply(rho, ordered_tilde_dual)))
        constraints = []
        for i in range(1, m + 1):
            constraints += [cp.sum(cp.multiply(rho[:i], answer[:i])) >= -lambd]
        prob = cp.Problem(objective, constraints)
        prob.solve()
        return answer.value




import os
import pickle
import torch
import pandas as pd
import numpy as np
from tqdm import tqdm

pd.set_option('display.max_rows', 2000)


class subtopic:
    def __init__(self, subtopic_id, subtopic):
        """
        Represents a subtopic of a query.

        :param subtopic_id: Unique identifier for the subtopic.
        :param subtopic: Text representation of the subtopic.
        """
        self.subtopic_id = subtopic_id
        self.subtopic = subtopic


class div_query:
    def __init__(self, qid, query, subtopic_id_list, subtopic_list):
        """
        Represents a diversity query for re-ranking search results.
        
        :param qid: Unique query identifier.
        :param query: Text of the query.
        :param subtopic_id_list: List of subtopic IDs associated with the query.
        :param subtopic_list: List of subtopic texts corresponding to the subtopic IDs.
        """
        self.qid = qid
        self.query = query
        self.subtopic_id_list = subtopic_id_list
        self.subtopic_list = []
        self.doc_list = []
        self.doc_score_list = []
        self.best_metric = 0
        self.stand_alpha_DCG = 0

        for index in range(len(subtopic_id_list)):
            t = subtopic(subtopic_id_list[index], subtopic_list[index])
            self.subtopic_list.append(t)

    def set_std_metric(self, m):
        """
        Sets the standard alpha-DCG metric for normalization.
        
        :param m: Standard alpha-DCG metric value.
        """
        self.stand_alpha_DCG = m

    def add_docs(self, doc_list):
        """
        Adds a list of documents to the query and initializes subtopic relevance tracking.
        
        :param doc_list: List of document identifiers.
        """
        self.doc_list = doc_list
        self.DOC_NUM = len(self.doc_list)
        init_data = np.zeros((len(doc_list), len(self.subtopic_list)), dtype=int)
        self.subtopic_df = pd.DataFrame(init_data, columns=self.subtopic_id_list, index=doc_list)

    def add_query_suggestion(self, query_suggestion):
        """
        Adds query suggestions related to the main query.
        
        :param query_suggestion: Suggested query string.
        """
        self.query_suggestion = query_suggestion

    def add_docs_rel_score(self, doc_score_list):
        """
        Adds relevance scores for the documents associated with the query.
        
        :param doc_score_list: List of relevance scores for documents.
        """
        self.doc_score_list = doc_score_list

    def get_test_alpha_nDCG(self, docs_rank):
        """
        Get the alpha_nDCG@20 for the input document list (for testing).
        
        :param docs_rank: Ordered list of document identifiers.
        :return: Alpha-nDCG score for the given ranking.
        """
        temp_data = np.zeros((len(docs_rank), len(self.subtopic_list)), dtype=int)
        temp_array = np.array(self.best_subtopic_df)
        metrics = []
        p = 0.5
        real_num = min(20, len(docs_rank))
        best_docs_index = []
        for index in range(real_num):
            result_index = self.best_docs_rank.index(docs_rank[index])
            best_docs_index.append(result_index)
            temp_data[index, :] = temp_array[result_index, :]
            if index == 0:
                score = np.sum(temp_data[index, :])
                metrics.append(score)
            else:
                r_ik = np.array([np.sum(temp_data[:index, s]) for s in range(temp_data.shape[1])], dtype=np.int64)
                t = np.power(p, r_ik)
                score = np.dot(temp_data[index, :], t) / np.log2(2 + index)
                metrics.append(score)
        ''' normalized by the stand alpha DCG '''
        if hasattr(self, 'stand_alpha_DCG') and self.stand_alpha_DCG > 0:
            try:
                alpha_nDCG = np.sum(metrics) / self.stand_alpha_DCG
            except:
                print('except np.sum =', np.sum(metrics), 'self.global_best_metric = ', self.global_best_metric)
        else:
            print('error! qid =', self.qid)
            alpha_nDCG = 0
        return alpha_nDCG
    

    def get_alpha_DCG(self, docs_rank, print_flag=False):
        """
        Computes the alpha-DCG for the input document list (for generating training samples)
        
        :param docs_rank: A list of document IDs representing the ranking order.
        :param print_flag: A boolean flag indicating whether to print intermediate computation results.
        :return: The computed alpha-DCG score for the given document ranking.
        """

        temp_data = np.zeros((len(docs_rank), len(self.subtopic_list)), dtype=int)
        temp_array = np.array(self.best_subtopic_df)
        metrics = []
        p = 0.5
        for index in range(len(docs_rank)):
            result_index = self.best_docs_rank.index(docs_rank[index])
            temp_data[index, :] = temp_array[result_index, :]
            if index == 0:
                score = np.sum(temp_data[index, :])
                metrics.append(score)
            else:
                r_ik = np.array([np.sum(temp_data[:index, s]) for s in range(temp_data.shape[1])], dtype=np.int64)
                t = np.power(p, r_ik)
                score = np.dot(temp_data[index, :], t) / np.log2(2 + index)
                metrics.append(score)
        if print_flag:
            print('self.best_gain = ', self.best_gain, 'sum(best_gain) = ', np.sum(self.best_gain), 'best_metric = ',
                  self.best_metric)
            print('test metrics = ', metrics, 'sum(metrics) = ', np.sum(metrics))
        '''get the total gain for the input document list'''
        alpha_nDCG = np.sum(metrics)
        return alpha_nDCG

    def get_best_rank(self, top_n=None, alpha=0.5):
        """
        Generates the best document ranking using a greedy selection strategy.
        
        :param top_n: The number of top documents to be selected (default: all available documents).
        :param alpha: A parameter controlling redundancy reduction (default: 0.5).
        :return: Updates class attributes with the best document ranking and associated gains.
        """

        p = 1.0 - alpha
        if top_n == None:
            top_n = self.DOC_NUM
        real_num = int(min(top_n, self.DOC_NUM))
        temp_data = np.zeros((real_num, len(self.subtopic_list)), dtype=int)
        temp_array = np.array(self.subtopic_df)
        best_docs_rank = []
        best_docs_rank_rel_score = []
        best_gain = []
        ''' greedy document selection '''
        for step in range(real_num):
            scores = []
            if step == 0:
                for index in range(real_num):
                    temp_score = np.sum(temp_array[index, :])
                    scores.append(temp_score)
                result_index = np.argsort(scores)[-1]
                gain = scores[result_index]
                docid = self.doc_list[result_index]
                doc_rel_score = self.doc_score_list[result_index]
                best_docs_rank.append(docid)
                best_docs_rank_rel_score.append(doc_rel_score)
                best_gain.append(scores[result_index])
                temp_data[0, :] = temp_array[result_index, :]
            else:
                for index in range(real_num):
                    if self.doc_list[index] not in best_docs_rank:
                        r_ik = np.array([np.sum(temp_data[:step, s]) for s in range(temp_array.shape[1])],
                                        dtype=np.int64)
                        t = np.power(p, r_ik)
                        temp_score = np.dot(temp_array[index, :], t)
                        scores.append(temp_score)
                    else:
                        scores.append(-1.0)
                result_index = np.argsort(scores)[-1]
                gain = scores[result_index]
                docid = self.doc_list[result_index]
                doc_rel_score = self.doc_score_list[result_index]
                if docid not in best_docs_rank:
                    best_docs_rank.append(docid)
                    best_docs_rank_rel_score.append(doc_rel_score)
                else:
                    print('document already added!')
                best_gain.append(scores[result_index] / np.log2(2 + step))
                temp_data[step, :] = temp_array[result_index, :]
        self.best_docs_rank = best_docs_rank
        self.best_docs_rank_rel_score = best_docs_rank_rel_score
        self.best_gain = best_gain
        self.best_subtopic_df = pd.DataFrame(temp_data, columns=self.subtopic_id_list, index=self.best_docs_rank)
        self.best_metric = np.sum(self.best_gain)


class div_dataset:
    def __init__(self, config):
        """
        Initializes the dataset object with file paths and configuration. 

        :param config: A dictionary containing configuration settings.
        """
        self.Best_File = os.path.join(config['data_dir'], 'div_query.data')
        self.Train_File = os.path.join(config['data_dir'], config['model'], 'listpair_train.data')
        if not os.path.exists(os.path.join(config['data_dir'], config['model'])):
            os.makedirs(os.path.join(config['data_dir'], config['model']))
        self.config = config

    def get_listpairs(self, div_query, context, top_n):
        """
        Generates list-pair samples
        
        :param div_query: The query object that contains the list of ranked documents.
        :param context: A list of previously considered documents in the context.
        :param top_n: The number of top-ranked documents to consider.
        :return: A list of generated samples, each containing metrics, positive/negative masks, and weights. 
        """
        best_rank = div_query.best_docs_rank
        metrics = []
        samples = []
        for index in range(len(best_rank)):
            if best_rank[index] not in context:
                metric = div_query.get_alpha_DCG(context + [best_rank[index]])
            else:
                metric = -1.0
            metrics.append(metric)
        ''' padding the metrics '''
        if len(metrics) < top_n:
            metrics.extend([0] * (top_n - len(metrics)))
        total_count = 0
        for i in range(len(best_rank)):
            ''' set a limit to the total sample number '''
            if total_count > 20:
                break
            count = 0
            for j in range(i + 1, len(best_rank)):
                ''' set a limit to sample number on the same context'''
                if count > 5:
                    break
                if metrics[i] < 0 or metrics[j] < 0 or metrics[i] == metrics[j]:
                    pass
                elif metrics[i] > metrics[j]:
                    count += 1
                    total_count += 1
                    positive_mask = torch.zeros(top_n)
                    negative_mask = torch.zeros(top_n)
                    weight = metrics[i] - metrics[j]
                    positive_mask[i] = 1
                    negative_mask[j] = 1
                    samples.append((metrics, positive_mask, negative_mask, weight))
                elif metrics[i] < metrics[j]:
                    count += 1
                    total_count += 1
                    positive_mask = torch.zeros(top_n)
                    negative_mask = torch.zeros(top_n)
                    weight = metrics[j] - metrics[i]
                    positive_mask[j] = 1
                    negative_mask[i] = 1
                    samples.append((metrics, positive_mask, negative_mask, weight))
        return samples

    def get_listpair_train_data(self, top_n=50):
        """
        Generates list-pair training samples using the top N relevant documents. 
        This function processes the best document ranks for each query, generates list-pair samples, and saves them to a file: listpair_train.data. 
        data_dict[qid] = [(metrics, positive_mask, negative_mask, weight),...]
        metrics, positive_mask and negative_mask are padding as tensors with length of top_n

        :param top_n: The number of top-ranked documents to use for generating the list-pairs.
        :return: Saves the generated list-pair training data into a file. 
        """
        qd = pickle.load(open(self.Best_File, 'rb'))
        train_dict = {}
        for qid in tqdm(qd, desc="Gen Train Data"):
            temp_q = qd[qid]
            result_list = []
            real_num = int(min(top_n, temp_q.DOC_NUM))
            for i in range(real_num):
                listpair_data = self.get_listpairs(temp_q, temp_q.best_docs_rank[:i], top_n)
                if len(listpair_data) > 0:
                    result_list.extend(listpair_data)
            train_dict[str(qid)] = result_list
        pickle.dump(train_dict, open(self.Train_File, 'wb'), True)



import os
import gzip
import time
import torch
import pickle
import argparse
import numpy as np
import sys
import json

from torch.nn.utils import clip_grad_norm_
from tqdm import tqdm
from sklearn.model_selection import KFold
from torch.utils.data import Dataset, DataLoader

from ..postprocessing_model.DALETOR import DALETOR
from ..utils.loss import ndcg_loss
from ..post_evaluator import evaluate_test_qids_DALETOR


class TrainDataset(Dataset):
    def __init__(self, train_list):
        """
        A PyTorch Dataset class for handling training data, ensuring they are properly prepared for model training.

        :param train_list: List of training samples containing input tensors and features
        """
        self.data = train_list

    def __len__(self):
        return len(self.data)

    def __getitem__(self, idx):
        X = self.data[idx][0].clone().detach().float()
        rel_feat = self.data[idx][1].clone().detach().float()
        div_feat = self.data[idx][2].clone().detach().float()
        X.requires_grad=False
        rel_feat.requires_grad=False
        div_feat.requires_grad=False
        return X, rel_feat, div_feat


def get_train_loader(fold, train_ids, config):
    """
    DataLoader for training data loading and processing.

    :param fold: Current fold number in cross-validation
    :param train_ids: List of training query IDs to load
    :param config: Dictionary containing configuration parameters
    :return: DataLoader object containing the processed training data
    """
    data_list = []
    data_dir=os.path.join(config['data_dir'], config['model'], 'train/')
    doc_tensor_dict={}
    for qid in tqdm(train_ids,desc='load train data', ncols=80):
        file_path = os.path.join(data_dir,str(qid)+'.pkl.gz')
        with gzip.open(file_path,'rb') as f:
            try:
                sample_dict = pickle.load(f)
            except EOFError:
                continue
        data_list.extend(sample_dict[str(qid)])
    train_dataset=TrainDataset(data_list)
    loader=DataLoader(
        dataset=train_dataset,
        batch_size=config['batch_size'],
        shuffle=False,
        num_workers=4,
        pin_memory=True,
    )
    return loader


def get_test_dataset(fold, test_qids, config):
    """
    Loads and processes test data for model evaluation.

    :param fold: Current fold number in cross-validation
    :param test_qids: List of test query IDs to load
    :param config: Dictionary containing configuration parameters
    :return: Dictionary containing processed test data organized by query ID
    """
    data_dir = os.path.join(config['data_dir'], config['model'], 'test/')
    test_dataset = {}
    for qid in tqdm(test_qids, desc="load test data",ncols=80):
        file_path = os.path.join(data_dir,str(qid)+'.pkl.gz')
        with gzip.open(file_path,'rb') as f:
            try:
                temp_test_dict=pickle.load(f)
            except EOFError:
                continue
        test_dataset[str(qid)]=temp_test_dict[str(qid)]
    return test_dataset


def DALETOR_run(config):
    """
    Executes the complete training and evaluation pipeline for the DALETOR model.
    This function performs 5-fold cross-validation training, including model initialization,
    training loop execution, periodic evaluation, learning rate adjustment, and model checkpointing.
    It tracks the best performing model for each fold and saves relevant metrics and model states.

    :param config: Dictionary containing model configuration parameters
    """
    if not os.path.exists(os.path.join(config['model_save_dir'], config['model'])):
        os.makedirs(os.path.join(config['model_save_dir'], config['model']))
    all_qids=np.load(os.path.join(config['data_dir'], 'all_qids.npy'))
    qd=pickle.load(open(os.path.join(config['data_dir'], 'div_query.data'),'rb'))
    
    final_metrics=[]
    best_model_list=[]
    fold_time=0
    
    test_qids_list=[]
    for train_ids, test_ids in KFold(5).split(all_qids):
        fold_time+=1
        max_metric = 0
        train_ids.sort()
        test_ids.sort()
        train_qids=[str(all_qids[i]) for i in train_ids]
        test_qids=[str(all_qids[i]) for i in test_ids]
        test_qids_list.append(test_qids)

        train_data_loader = get_train_loader(fold_time, train_qids, config)
        test_dataset_dict = get_test_dataset(fold_time, test_qids, config)

        model = DALETOR(config['dropout'])
        if torch.cuda.is_available():
            model=model.cuda()
        opt = torch.optim.Adagrad(model.parameters(), lr=config['learning_rate'])
        params = list(model.parameters())
        if fold_time == 1:
            print('model={}'.format(model))
            print('len params={}'.format(len(params)))
            for param in params:
                print('{}'.format(param.size()))
            n_params = sum([p.numel() for p in model.parameters() if p.requires_grad])
            print('* number of parameters: %d' % n_params)

        all_steps = len(train_data_loader)
        patience = 0
        
        for epoch in range(config['epoch']):
            print('Start Training...')
            epoch_step=0
            model.train()
            
            for step, train_data in enumerate(tqdm(train_data_loader, desc='BATCH', ncols=80)):
                X, rel_feat, div_feat = train_data
                if torch.cuda.is_available():
                    X = X.cuda()
                    rel_feat = rel_feat.cuda()
                    div_feat = div_feat.cuda()
                    
                score = model.fit(X, rel_feat, True)
                loss = ndcg_loss(score, div_feat)

                opt.zero_grad()
                loss.backward()
                clip_grad_norm_(model.parameters(), max_norm=1)
                opt.step()

                epoch_step+=1
                if (step + 1) % config['eval_steps'] == 0:
                    model.eval()
                    metrics = []
                    for qid in test_qids:
                        metric = evaluate_test_qids_DALETOR(model, test_dataset_dict[str(qid)], qd[str(qid)], 'metric')
                        metrics.append(metric)
                    avg_alpha_NDCG = np.mean(metrics)

                    if max_metric < avg_alpha_NDCG:
                        max_metric = avg_alpha_NDCG
                        print('max avg_alpha_NDCG updated: {}'.format(max_metric))
                        model_filename = os.path.join(config['model_save_dir'], config['model'], 'TOTAL_EPOCH_' + str(config['epoch']) + '_FOLD_' + str(fold_time) + '_EPOCH_' + str(epoch) + '_LR_' + str(config['learning_rate']) + '_BATCHSIZE_' + str(
                        config['batch_size']) + '_DROPOUT_' + str(config['dropout']) + '_' + str(config['embedding_type']) + '.pickle')
                        torch.save(model.state_dict(),model_filename)
                        best_model = model_filename
                        patience = 0
                    else: 
                        patience += 1
                    model.train()

                    if epoch > 0 and patience > 2:
                        new_lr = 0.0
                        for param_group in opt.param_groups:
                            param_group['lr'] = param_group['lr'] * 0.5
                            new_lr = param_group['lr']
                        patience = 0

            model.eval()
            metrics = []
            for qid in test_qids:
                metric = evaluate_test_qids_DALETOR(model, test_dataset_dict[str(qid)], qd[str(qid)], 'metric')
                metrics.append(metric)
            avg_alpha_NDCG = np.mean(metrics)
            if max_metric < avg_alpha_NDCG:
                max_metric = avg_alpha_NDCG
                model_filename = os.path.join(config['model_save_dir'], config['model'], 'TOTAL_EPOCH_' + str(config['epoch']) + '_FOLD_' + str(fold_time) + '_EPOCH_' + str(epoch) + '_LR_' + str(config['learning_rate']) + '_BATCHSIZE_' + str(
                        config['batch_size']) + '_DROPOUT_' + str(config['dropout']) + '_' + str(config['embedding_type']) + '.pickle')
                torch.save(model.state_dict(),model_filename)
                best_model = model_filename
            if epoch == (config['epoch']-1):
                final_metrics.append(max_metric)
                best_model_list.append(best_model)
    with open(os.path.join(config['model_save_dir'], config['model'], 'fold_qid.json'), 'w') as f:
        json.dump(test_qids_list, f)
    print('alpha-nDCG = {}, best model = {}'.format(sum(final_metrics)/len(final_metrics), best_model_list))



import torch
import torch.nn as nn
import numpy as np
from .abstract_model import AbstractBaseModel
from .layers import TransformerEncoder

r"""
GRU4Rec
################################################

Reference:
    Wang-Cheng Kang et al. "Self-Attentive Sequential Recommendation." in ICDM 2018.

"""

class SASRec(AbstractBaseModel):
    def __init__(self, config):
        super().__init__(config)
        # self.user_embedding = nn.Embedding(num_users, embedding_dim)
        # self.item_embedding = nn.Embedding(num_items, embedding_dim)
        self.type = ["sequential"]
        self.IR_type = ["retrieval", "ranking"]

        self.n_layers = config["n_layers"]
        self.n_heads = config["n_heads"]
        self.hidden_size = config["embedding_size"]  # same as embedding_size
        self.inner_size = config[
            "inner_size"
        ]  # the dimensionality in feed-forward layer
        self.hidden_dropout_prob = config["hidden_dropout_prob"]
        self.attn_dropout_prob = config["attn_dropout_prob"]
        self.hidden_act = config["hidden_act"]
        self.layer_norm_eps = float(config["layer_norm_eps"])

        self.initializer_range = config["initializer_range"]

        self.history_length = config['history_length']

        self.position_embedding = nn.Embedding(self.history_length, self.hidden_size)
        self.trm_encoder = TransformerEncoder(
            n_layers=self.n_layers,
            n_heads=self.n_heads,
            hidden_size=self.hidden_size,
            inner_size=self.inner_size,
            hidden_dropout_prob=self.hidden_dropout_prob,
            attn_dropout_prob=self.attn_dropout_prob,
            hidden_act=self.hidden_act,
            layer_norm_eps=self.layer_norm_eps,
        )

        self.LayerNorm = nn.LayerNorm(self.hidden_size, eps=self.layer_norm_eps)
        self.dropout = nn.Dropout(self.hidden_dropout_prob)

        self.sigmoid = nn.Sigmoid()
        self.loss = nn.CrossEntropyLoss(reduction='none')
        self.apply(self._init_weights)


    # def forward(self, user_ids, item_ids):
    #     user_embeds = self.user_embedding(user_ids)
    #     item_embeds = self.item_embedding(item_ids)
    #     dot_product = (user_embeds * item_embeds).sum(1)
    #     return self.sigmoid(dot_product)


    def compute_loss(self, interaction):

        #user = interaction['user_ids']
        #history_ids = interaction['history_ids']
        pos_items = interaction['item_ids']
        user_emb = self.get_user_embedding(interaction)
        test_item_emb = self.item_embedding.weight
        logits = torch.matmul(user_emb, test_item_emb.transpose(0, 1))
        loss = self.loss(logits, pos_items)
        return loss

    def get_user_embedding(self, user_dict):
        users = user_dict['history_ids']
        item_seq_emb = self.item_embedding(users)
        ids = torch.arange(
            self.history_length, dtype=torch.long, device=users.device
        )
        position_ids = ids.repeat(users.size(0), 1)
        position_embedding = self.position_embedding(position_ids)
        input_emb = item_seq_emb + position_embedding
        input_emb = self.LayerNorm(input_emb)
        input_emb = self.dropout(input_emb)

        trm_output = self.trm_encoder(
            input_emb, torch.ones_like(ids).to(input_emb.device), output_all_encoded_layers=True
        )
        output = trm_output[-1]
        seq_output = output[:,-1,:]
        #seq_output = self.gather_indexes(gru_output, self.config['history_length'] - 1)
        return seq_output



    def forward(self, user_dict, item_id):
        ###here user_id denotes the historical ids
        user_embeds = self.get_user_embedding(user_dict)
        item_embeds = self.item_embedding(item_id)
        dot_product = (user_embeds * item_embeds).sum(1)
        return self.sigmoid(dot_product)




from __future__ import division
import numpy as np
from scipy.special import softmax
from sklearn.metrics import pairwise


def dist_prototype(X, v, alpha):
    """
    Compute the Euclidean distance between data points and prototypes with a weighted sum.

    :param X : numpy.ndarray
        A 2D array where each row represents a data point.
    :param v : numpy.ndarray
        A 2D array where each row represents a prototype.
    :param alpha : numpy.ndarray
        A 1D array of weights for the features.

    :return : numpy.ndarray
        A 1D array of distances between the data points and prototypes.
    """
    dists = np.sum(alpha * (np.array(X[:, np.newaxis, :]) - np.array(v[np.newaxis, :, :])) ** 2, axis=2)
    return np.sqrt(dists)


def dist_pairwise_group(Xg1, Xg2, alpha):
    """
    Compute the pairwise distances between two sets of data points with feature weights.

    :param Xg1 : numpy.ndarray
        A 2D array representing the first set of data points.
    :param Xg2 : numpy.ndarray
        A 2D array representing the second set of data points.
    :param alpha : numpy.ndarray
        A 1D array of weights for the features.

    :return : numpy.ndarray
        A 2D array containing the pairwise distances between points from Xg1 and Xg2.
    """
    dists = np.sum(alpha * (Xg1[:, np.newaxis, :] - Xg2[np.newaxis, :, :]) ** 2, axis=2)
    return dists


def compute_U_nk(dists):
    """
    Compute the membership matrix based on distances using the softmax function.

    :param dists : numpy.ndarray
        A 2D array of distances between data points and prototypes.

    :return : numpy.ndarray
        A 2D array representing the membership matrix.
    """
    M = softmax(-dists, axis=1)
    return M


def replaceNaNwithMean(X):
    """
    Replaces NaN values in the dataset with the column mean.

    :param X : numpy.ndarray
        A 2D array with potential NaN values.

    :return : numpy.ndarray
        A 2D array with NaN values replaced by the column mean.
    """
    # Obtain mean of columns as you need, nanmean is just convenient.
    col_mean = np.nanmean(X, axis=0)
    # - guarding against a column where all values are NaN
    col_mean[np.isnan(col_mean)] = 0

    # Find indices that need replacement
    inds = np.where(np.isnan(X))

    # Replace NaN values with column means
    X[inds] = np.take(col_mean, inds[1])

    return X


def compute_X_hat(X, params, k, alpha=False):
    """
    Compute the approximated data X_hat using the given parameters.

    :param X : numpy.ndarray
        A 2D array of data points to approximate.
    :param params : numpy.ndarray
        A 1D array containing the model parameters.
    :param k : int
        The number of prototypes.
    :param alpha : bool, optional, default=False
        Whether to use feature-specific weights.

    :return : tuple
        A tuple containing the approximated data X_hat and the membership matrix U_nk.
    """
    N, M = X.shape

    if alpha:
        prototypes = np.matrix(params[M:]).reshape((k, M))
        alpha = params[:M]
    else:
        prototypes = np.matrix(params[k:]).reshape((k, M))
        alpha = np.ones(M)

    distances_prototypes = dist_prototype(X, prototypes, alpha)
    U_nk = compute_U_nk(distances_prototypes)
    X_hat = np.matmul(U_nk, prototypes)

    return np.array(X_hat), U_nk


def get_xhat_y_hat(X, params, k):
    """
    Compute the approximated data X_hat and predicted values Y_hat.

    :param X : numpy.ndarray
        A 2D array of data points to approximate.
    :param params : numpy.ndarray
        A 1D array containing the model parameters.
    :param k : int
        The number of prototypes.

    :return : tuple
        A tuple containing the membership matrix U_nk, the approximated data X_hat,
        and the predicted values Y_hat.
    """
    X_hat, U_nk = compute_X_hat(X, params, k)

    w = params[:k]
    Y_hat = np.clip(
        np.matmul(U_nk, w.reshape((-1, 1))),
        np.finfo(float).eps,
        1.0 - np.finfo(float).eps
    )
    return U_nk, X_hat, Y_hat


def compute_euclidean_distances(X, nonsensitive_column_indices):
    """
    Compute the Euclidean distance matrix based on the non-sensitive columns of the data.

    :param X : numpy.ndarray
        A 2D array of data points.
    :param nonsensitive_column_indices : list
        A list of indices identifying the non-sensitive columns.

    :return : numpy.ndarray
        A 2D array representing the pairwise Euclidean distances between non-sensitive data points.
    """
    X_non_sensitive = X[:, nonsensitive_column_indices]
    if len(X_non_sensitive.shape) > 2:
        X_non_sensitive = X_non_sensitive.squeeze()
    D_X_F = pairwise.euclidean_distances(X_non_sensitive, X_non_sensitive)
    return D_X_F




from tqdm import tqdm



class Prompt_Constructer(object):
    def __init__(self, config):
        self.item_feature_list = ['id','title','publisher']
        self.item_domain = config['item_domain']
        self.fair_prompt = config['fair_prompt']
        self.history_behavior_field = 'history_behaviors'
        self.item_candidate_field = 'items'
        self.pos_length_field = 'pos_length'

    def construct_inter_dict(self, input_file, iid2title, iid2cate):
        """
        Constructs an interaction dictionary from the given input data.

        This method processes an input file, typically a DataFrame, to create a dictionary representation of user interactions.
        It extracts user history, candidate items, positive and negative samples for recommendation tasks. The method also
        enriches item information using provided mappings for item titles and categories.


        :param input_file (pandas.DataFrame): The DataFrame containing user interaction data with necessary fields.
        :param iid2title (dict): A mapping of item IDs to their respective titles.
        :param iid2cate (dict): A mapping of item IDs to their respective categories.

        :return: data_dict (dict): A dictionary where each key is a user ID and the value is another dictionary containing:
            - 'history_items' (list[dict]): List of dictionaries representing historical items interacted by the user,
              each with 'id', 'title', and 'publisher' keys.
            - 'item_candidates' (list): List of candidate items for the user.
            - 'positive_items' (list): Subset of 'item_candidates' up to the positive sample length, indicating observed interactions.
            - 'negative_items' (list): Remaining items in 'item_candidates', considered as negative samples.
        """
        data_dict = {}
        print(input_file.head)
        for index, row in input_file.iterrows():
            userid = row['user_id']
            history_itemid_list = eval(row[self.history_behavior_field])
            item_candidate = eval(row[self.item_candidate_field])
            pos_len = int(row[self.pos_length_field])
            data_dict[userid] = {
                'history_items': [],
                'item_candidates': item_candidate,
                'positive_items': item_candidate[:pos_len],
                'negative_items': item_candidate[pos_len:]
            }
            for item_id in history_itemid_list:
                item_feat_dict = {'id': item_id, # item id
                                  'title': iid2title.get(item_id, 'unknown'),
                                  'publisher': iid2cate.get(item_id, 'unknown')
                                  }
                data_dict[userid]['history_items'].append(item_feat_dict)
        return data_dict

    def construct_prompt(self, input_file, iid2title, iid2pid):
        """
        Constructs a prompt dataset from the given input.

        :param input_file (str): The path to the input file containing necessary data for constructing the prompts.
        :param iid2title (dict): A dictionary mapping identifiers to their respective titles, enhancing context in prompts.
        :param iid2pid (dict): A dictionary associating identifiers with parent identifiers, adding hierarchical information.

        :return: str: A JSON string representing the constructed dataset, formatted suitably for use as prompts.

        """
        # print(cates_name)
        data_dict = self.construct_inter_dict(input_file, iid2title, iid2pid)
        prompt_dataset_json = self.data_to_json(data_dict)
        return prompt_dataset_json

    def data_to_json(self, data_dict):
        """
        Converts a dictionary of user data into a JSON formatted list for recommendation tasks.

        :param data_dict (Dict[str, Dict]): A dictionary where keys are user identifiers and values are dictionaries containing
          'history_items' (a list of items viewed by the user), 'item_candidates' (a list of candidate items for recommendation),
          'positive_items' (items liked by the user, if any), and 'negative_items' (items not liked by the user, if any).

        :return: List[Dict]: A list of dictionaries, each representing a recommendation task with structured information including
          instruction, input context based on user history, and candidate items for generating recommendations.
        """
        json_list = []
        for user, feats in tqdm(data_dict.items()):
            history = f"The user has viewed the following {self.item_domain}s before, with features as: "
            for item in feats['history_items']:
                # history += f"Item id: {item['item_id']}:"
                for feat in self.item_feature_list:
                    history += f"Item {feat}: {item[feat]},"
                history += '\n'
            # target_item = feats['positive_items']
            # target_movie_str = "" + str(target_item) + ""
            instruction = f"You are a {self.item_domain} recommender. Given a list of {self.item_domain} the user has clicked before, please recommend a item that the user may like."
            instruction += self.fair_prompt if self.fair_prompt else f""
            json_list.append({
                "instruction": instruction,
                "input": f"{history}",
                # "output": target_movie_str,
                'item_candidates': feats['item_candidates'],
                'positive_items': feats['positive_items'],
                'negative_items': feats['negative_items'],
                # "sensitiveAttribute": iid2pid[target_item]
            })
        return json_list


from sklearn.metrics import roc_auc_score
import numpy as np

##############ranking metrics #############################

def AUC_score(y_scores, y_true):
    r"""AUC_ (also known as Area Under Curve) is used to evaluate the two-class model, referring to
        the area under the ROC curve.

        .. _AUC: https://en.wikipedia.org/wiki/Receiver_operating_characteristic#Area_under_the_curve

        Note:
            This metric does not calculate group-based AUC which considers the AUC scores
            averaged across users. It is also not limited to k. Instead, it calculates the
            scores on the entire prediction results regardless the users. We call the interface
            in `scikit-learn`, and code calculates the metric using the variation of following formula.

        .. math::
            \mathrm {AUC} = \frac {{{M} \times {(N+1)} - \frac{M \times (M+1)}{2}} -
            \sum\limits_{i=1}^{M} rank_{i}} {{M} \times {(N - M)}}

        :math:`M` denotes the number of positive items.
        :math:`N` denotes the total number of user-item interactions.
        :math:`rank_i` denotes the descending rank of the i-th positive item.
    """
    auc_score = roc_auc_score(y_true, y_scores)
    return auc_score

def dcg(scores, k):
    """
    Calculate the Discounted Cumulative Gain (DCG) at rank k.
    """
    scores = np.array(scores)[:k]
    if scores.size:
        return np.sum(scores / np.log2(np.arange(2, scores.size + 2)))
    return 0.0


def NDCG(ranking_list, label_list, k):
    r"""NDCG_ (also known as normalized discounted cumulative gain) is a measure of ranking quality,
    where positions are discounted logarithmically. It accounts for the position of the hit by assigning
    higher scores to hits at top ranks.

    .. _NDCG: https://en.wikipedia.org/wiki/Discounted_cumulative_gain#Normalized_DCG

    .. math::
        \mathrm {NDCG@K} = \frac{1}{|U|}\sum_{u \in U} (\frac{1}{\sum_{i=1}^{\min (|R(u)|, K)}
        \frac{1}{\log _{2}(i+1)}} \sum_{i=1}^{K} \delta(i \in R(u)) \frac{1}{\log _{2}(i+1)})

    :math:`\delta(·)` is an indicator function.
    """
    sorted_label = np.sort(label_list)[::-1]
    label_dcg = dcg(sorted_label, k)
    ranking_dcg = dcg(ranking_list, k)
    ndcg = ranking_dcg/label_dcg

    return ndcg

def HR(ranking_list, label_list, k):
    r"""HR_ (also known as truncated Hit-Ratio) is a way of calculating how many 'hits'
        you have in an n-sized list of ranked items. If there is at least one item that falls in the ground-truth set,
        we call it a hit.

        .. _HR: https://medium.com/@rishabhbhatia315/recommendation-system-evaluation-metrics-3f6739288870

        .. math::
            \mathrm {HR@K} = \frac{1}{|U|}\sum_{u \in U} \delta(\hat{R}(u) \cap R(u) \neq \emptyset),

        :math:`\delta(·)` is an indicator function. :math:`\delta(b)` = 1 if :math:`b` is true and 0 otherwise.
        :math:`\emptyset` denotes the empty set.
    """
    sorted_label = np.sort(label_list)[::-1]
    hr = np.sum(ranking_list[:k])/np.sum(sorted_label[:k])
    return hr


def MRR(ranking_list, k):
    r"""The MRR_ (also known as Mean Reciprocal Rank) computes the reciprocal rank
        of the first relevant item found by an algorithm.

        .. _MRR: https://en.wikipedia.org/wiki/Mean_reciprocal_rank

        .. math::
           \mathrm {MRR@K} = \frac{1}{|U|}\sum_{u \in U} \frac{1}{\operatorname{rank}_{u}^{*}}

        :math:`{rank}_{u}^{*}` is the rank position of the first relevant item found by an algorithm for a user :math:`u`.
    """
    mrr = 0
    for index, i in enumerate(ranking_list[:k]):
        if i > 0:
            mrr = i/(index+1)
            break
    return mrr


##################fairness metric#######################

def reconstruct_utility(utility_list, weights, group_mask):
    """
        Reconstruct utility by re-weighting them and masking the utility of certain unused groups.
        :param utility_list: array for item/user utilities
        :param weights: array for item/user utilities weights
        :param group_mask: bool array for whether computed the group utilityes
        :return: re-constructed utility array
    """

    if not weights:
        weights = np.ones_like(utility_list)

    utility_list = np.array(utility_list)
    weights = np.array(weights)
    weighted_utility = utility_list * weights

    if group_mask:
        weighted_utility = mask_utility(weighted_utility, group_mask)


    return np.array(weighted_utility)

def mask_utility(utility, group_mask):
    """
    Mask the utility values based on the provided group mask.

    This function filters out the utility values where the corresponding
    group mask element is zero, effectively removing them from the output.


    :param utility: array for item/user utilities
    :param group_mask: bool array for whether computed the group utilityes
    :return: masked utility array
    """


    masked_utility = []
    for i, m in enumerate(group_mask):
        if m == 0:
            masked_utility.append(utility[i])

    return np.array(masked_utility)

def MMF(utility_list, ratio=0.5, weights=None, group_mask = None):
    """
    Calculate the Max-min Fairness (MMF) index based on a given utility list.

    Parameters
    :param utility: array-like
        A list or array representing the utilities of resources or users.
    :param ratio: float, optional
        The fraction of the minimum utilities to consider for the MMF calculation. Defaults to 0.5.
    :param ratio: float, optional
        The fraction of the minimum utilities to consider for the MMF calculation. Defaults to 0.5.
    :param ratio: float, optional
        The fraction of the minimum utilities to consider for the MMF calculation. Defaults to 0.5.
    :param weights : array-like, optional
        An optional list or array of weights corresponding to each utility in `utility_list`.
        If provided, utilities are multiplied by their respective weights before sorting.
        Defaults to None, implying equal weighting.
    :param group_mask : array-like, optional
        An optional list or array used to selectively apply weights. If provided, it must have the same length as
        `utility_list` and `weights`. Defaults to None, indicating no group-based weighting.

    :return: The computed MMF index, representing the fairness of the allocation.
    """
    weighted_utility = reconstruct_utility(utility_list, weights, group_mask)

    MMF_length = int(ratio * len(utility_list))
    utility_sort = np.sort(weighted_utility)

    mmf = np.sum(utility_sort[:MMF_length])/np.sum(weighted_utility)

    return mmf

def MinMaxRatio(utility_list, weights=None, group_mask = None):
    """
    This function computes the minimum-to-maximum ratio of a list of utilities, optionally weighted and grouped.

    :param utility_list (list of float): A list containing numerical utility values.
    :param weights (list of float, optional): A list of weights corresponding to the utilities in utility_list. If provided,
      each utility is multiplied by its respective weight. If None, all utilities are considered with equal weight.
    :param group_mask (list of int or bool, optional): A mask indicating groups within the utility_list. If provided, it must
      be of the same length as utility_list. Groups are defined by consecutive True or 1 values. If None, no grouping is applied.

    :return: float: The computed minimum-to-maximum ratio of the (weighted) utilities.
    """
    weighted_utility = reconstruct_utility(utility_list, weights, group_mask)
    return np.min(weighted_utility) / np.max(weighted_utility)


def Gini(utility_list, weights=None, group_mask = None):
    """
    This function computes the Gini coefficient, a measure of statistical dispersion intended to represent income inequality within a nation or social group.
    The Gini coefficient is calculated based on the cumulative distribution of values in `utility_list`, which can optionally be weighted and masked.

    :param utility_list: array_like
        A 1D array representing individual utilities. The utilities are used to compute the Gini coefficient.
    :param weights: array_like, optional
        A 1D array of weights corresponding to `utility_list`. If provided, each utility value is multiplied by its respective weight before calculating the Gini coefficient. Defaults to None, implying equal weighting.
    :param group_mask: array_like, optional
        A 1D boolean array used to selectively include elements from `utility_list`. If provided, only the elements where the mask is True are considered in the calculation. Defaults to None, meaning all elements are included.

    :return: float: The computed Gini coefficient, ranging from 0 (perfect equality) to 1 (maximal inequality).
    """
    weighted_utility = reconstruct_utility(utility_list, weights, group_mask)

    values = np.sort(weighted_utility)
    n = len(values)

    # gini compute
    cumulative_sum = np.cumsum(values)
    gini = (n + 1 - 2 * (np.sum(cumulative_sum) / cumulative_sum[-1])) / n

    return gini

def Entropy(utility_list, weights=None, group_mask = None):
    """
    Calculate the entropy of a distribution given by `utility_list`, optionally
    weighted by `weights` and filtered by `group_mask`. Entropy measures the
    disorder or uncertainty in the distribution.

    :param utility_list: list or array-like
        A list or array representing utility values for each item.
    :param weights: list or array-like, optional
        A list or array of weights corresponding to each utility value. If not provided,
        all utilities are considered equally weighted. Defaults to None.
    :param group_mask : list or array-like, optional
        A boolean mask indicating which utilities to include in the calculation.
        If not provided, all utilities are included. Defaults to None.

    :return: float: The calculated entropy of the (potentially weighted and masked) distribution.

    Notes
    - Entropy is calculated as H = -sum(p * log2(p)), where p is the probability of each event.
    - Probabilities are normalized to ensure their sum equals 1.
    - To avoid taking the log of zero, a small constant (1e-9) is added to each probability before calculating the entropy.
    """
    weighted_utility = reconstruct_utility(utility_list, weights, group_mask)

    values = np.array(weighted_utility)
    values = values / np.sum(values)

    # H = - sum(p * log2(p))
    # avoid 0 case
    entropy_value = -np.sum(values * np.log2(values + 1e-9))
    return entropy_value






import os
import numpy as np
from .Abstract_Reranker import Abstract_Reranker
from tqdm import tqdm,trange
import torch

r"""
FairSync, a post-processing method to deal with minimum guarantee of different groups in retrieval process. 
Note that in this toolkit, in order to compare the performance fairly, we simplify it as the post-processing for ranking scores, 
which has the same effect of original paper.
 
The original FairSync is implemented in https://github.com/XuChen0427/FairSync/ using the faiss retrieval toolkit.

@inproceedings{Xu24FairSync,
author = {Xu, Chen and Xu, Jun and Ding, Yiming and Zhang, Xiao and Qi, Qi},
title = {FairSync: Ensuring Amortized Group Exposure in Distributed Recommendation Retrieval},
year = {2024},
isbn = {9798400701719},
publisher = {Association for Computing Machinery},
address = {New York, NY, USA},
url = {https://doi.org/10.1145/3589334.3645413},
doi = {10.1145/3589334.3645413},
booktitle = {Proceedings of the ACM Web Conference 2024},
pages = {1092–1102},
numpages = {11},
keywords = {distributed retrieval, minimum exposures, recommender system},
location = {Singapore, Singapore},
series = {WWW '24}
}

"""

class FairSync(Abstract_Reranker):
    def __init__(self, config, weights = None):
        super().__init__(config, weights)


    def rerank(self, ranking_score, k):
        ## its parameters
        user_size = len(ranking_score)
        mu_t = torch.zeros(self.group_num, requires_grad=True)
        rerank_list = []
        minimum_exposure = self.config['minimum_exposure']
        optimizer = torch.optim.Adam([mu_t, ], lr=self.config['learning_rate'])

        for u in trange(user_size):
            loss = 0.0
            query = mu_t.cpu().detach().numpy()
            #print(query)
            rel = ranking_score[u,:] - np.matmul(self.M, query)
            result_item = np.argsort(rel)[::-1]
            result_item = result_item[:k]
            rerank_list.append(result_item)
            for i in result_item:
                pid = np.argmax(self.M[i])
                loss = loss + torch.tensor(ranking_score[u,i]) - mu_t[pid]
            #B_l = B_l + np.sum(self.M[result_item,:],axis=0,keepdims=False)
            loss += (torch.sum(minimum_exposure * mu_t) + torch.max(mu_t) * (user_size * k - minimum_exposure * self.config['group_num'])) / user_size
            optimizer.zero_grad()
            # loss =
            loss.backward()
            optimizer.step()

        return rerank_list



from abc import ABC, abstractmethod
import torch.nn as nn

class BasePostProcessModel(nn.Module, ABC):
    def __init__(self, dropout=0.1):
        r"""
        A base class for post-processing supervised models.

        :param dropout: The dropout rate used in the model to reduce overfitting. Default is 0.1.
        """
        super(BasePostProcessModel, self).__init__()
        self.dropout = dropout
    
    @abstractmethod
    def fit(self):
        r"""
        Train the post-processing model.
        """
        pass

class BasePostProcessUnsupervisedModel(ABC):
    def __init__(self, top_k):
        r"""
        A base class for unsupervised post-processing models.

        :param top_k: The number of top-ranked items to retain after reranking.
        """
        super(BasePostProcessUnsupervisedModel, self).__init__()
        self.top_k = top_k
    
    @abstractmethod
    def rerank(self):
        r"""
        Reranks documents based on an unsupervised criterion.
        """
        pass


import os
import json
import pickle
import numpy as np

from ..utils.utils import pkl_load, get_metrics_20
from .base import BasePostProcessUnsupervisedModel

'''
@inproceedings{dang2012diversity,
  title={Diversity by proportionality: an election-based approach to search result diversification},
  author={Dang, Van and Croft, W Bruce},
  booktitle={Proceedings of the 35th international ACM SIGIR conference on Research and development in information retrieval},
  pages={65--74},
  year={2012}
}
'''

class PM2(BasePostProcessUnsupervisedModel):
    def __init__(self, top_k=20):
        """ 
        Initializes the PM2 for diversified document ranking.

        :param top_k: The maximum number of documents to be ranked. Default is 20.
        """
        super().__init__(top_k)
    
    def rerank(self, config):
        """ 
        Re-ranks documents using the PM-2 algorithm based on BM25 scores and query diversification.

        :param config: A dictionary containing configuration parameters.
        """

        qd = pickle.load(open(os.path.join(config['data_dir'], 'div_query.data'), 'rb'))
        bm25_dict = pkl_load(os.path.join(config['data_dir'], config['model'], 'bm25_scores.pkl'))

        output_dir = os.path.join(config['tmp_dir'], config['model'])
        if not os.path.exists(output_dir):
            os.makedirs(output_dir)

        output_txt = os.path.join(output_dir, 'output_best.txt')
        csv_path = os.path.join(output_dir, 'result.csv')

        for qid in qd:
            doc_ranking = self.calculate_pm2_score(qid, bm25_dict, config['lambda'])

            if doc_ranking == []:
              doc_ranking = qd[qid].doc_list[:self.top_k]
            judge_f = open(os.path.join(output_dir, str(qid)+'.txt'), 'w')
            for i in range(len(doc_ranking)):
                doc_ranking_score = len(doc_ranking) - i
                judge_f.write(str(qid) + ' Q0 ' + doc_ranking[i] + ' ' + str(i + 1) + ' ' + str(doc_ranking_score) + ' indri\n')
            judge_f.close()

        command = 'cat '+output_dir+'/* > '+output_txt
        os.system(command)
        command = './search/eval/clueweb09/ndeval ./search/eval/clueweb09/2009-2012.diversity.ndeval.qrels '+output_txt+' >' + str(csv_path)
        os.system(command)

        alpha_nDCG_20, NRBP_20, ERR_IA_20, S_rec_20 = get_metrics_20(csv_path)
        print('alpha_nDCG@20_std = {}, NRBP_20 = {}, ERR_IA_20 = {}, S_rec_20 = {}'.format(alpha_nDCG_20, NRBP_20, ERR_IA_20, S_rec_20))


    def calculate_pm2_score(self, qid, bm25_dict, _lambda=0.5):
        """ 
        Implements the PM-2 algorithm for diversified document ranking.

        :param qid: Query ID for which documents are being ranked.
        :param bm25_dict: Dictionary containing BM25 scores for both original and suggestion queries.
        :param _lambda: A trade-off parameter (0 to 1) that balances relevance and diversity. 
                        A higher value favors relevance, while a lower value emphasizes diversity.
        :return: A list of selected document IDs in diversified ranking order.
        """
        # Initialize variables
        doc_list = list(bm25_dict[qid][0].keys())  # Available documents
        suggestion_scores = bm25_dict[qid][1]  # Query suggestions' BM25 scores
        num_suggestions = len(suggestion_scores)
        if num_suggestions == 0:
          return []
        
        # Initialize seat allocation for each aspect (query suggestion)
        s = np.zeros(num_suggestions)
        
        # Initialize selected documents set and remaining documents
        S = []  # Selected documents
        R = doc_list.copy()  # Remaining documents
        
        # Main loop for selecting documents
        while len(S) < self.top_k and R:
            # Calculate quotient for each aspect
            quotient = np.zeros(num_suggestions)
            for i in range(num_suggestions):
                # v_i is assumed to be 1 for all aspects (uniform importance)
                v_i = 1
                quotient[i] = v_i / (2 * s[i] + 1)
                
            # Find aspect with maximum quotient
            i_star = np.argmax(quotient)
            
            # Calculate score for each remaining document
            best_score = float('-inf')
            d_star = None
            
            for d_j in R:
                doc_idx = doc_list.index(d_j)
                
                # Calculate first term (relevance to chosen aspect)
                first_term = _lambda * quotient[i_star] * suggestion_scores[i_star][doc_idx]
                
                # Calculate second term (relevance to other aspects)
                second_term = (1 - _lambda) * sum(
                    quotient[i] * suggestion_scores[i][doc_idx]
                    for i in range(num_suggestions)
                    if i != i_star
                )
                
                # Combined score
                score = first_term + second_term
                
                if score > best_score:
                    best_score = score
                    d_star = d_j
            
            if d_star is None:
                break
                
            # Update sets
            S.append(d_star)
            R.remove(d_star)
            
            # Update seat allocations
            doc_idx = doc_list.index(d_star)
            total_relevance = sum(suggestion_scores[j][doc_idx] for j in range(num_suggestions))
            
            for i in range(num_suggestions):
                if total_relevance > 0:  # Avoid division by zero
                    s[i] += suggestion_scores[i][doc_idx] / total_relevance
        
        return S


import os
import numpy as np
from .Abstract_Reranker import Abstract_Reranker
from tqdm import tqdm,trange

r"""
min-regularizer, a heuristic baseline aim to support the worst-off providers. It is proposed by

@inproceedings{xu2023p,
  title={P-MMF: Provider max-min fairness re-ranking in recommender system},
  author={Xu, Chen and Chen, Sirui and Xu, Jun and Shen, Weiran and Zhang, Xiao and Wang, Gang and Dong, Zhenhua},
  booktitle={Proceedings of the ACM Web Conference 2023},
  pages={3701--3711},
  year={2023}
}

"""

class min_regularizer(Abstract_Reranker):
    def __init__(self, config, weights = None):
        super().__init__(config, weights)


    def rerank(self, ranking_score, k):
        ## its parameters
        user_size = len(ranking_score)
        B_l = np.zeros(self.group_num)
        lambd = self.config['lambda']
        rerank_list = []

        for u in trange(user_size):
            minimax_reg = lambd * np.matmul(self.M,(-B_l + np.min(B_l))/(self.weights * user_size))
            rel = ranking_score[u,:] + minimax_reg
            result_item = np.argsort(rel)[::-1]
            result_item = result_item[:k]
            rerank_list.append(result_item)
            B_l = B_l + np.sum(self.M[result_item,:],axis=0,keepdims=False)

        return rerank_list



import numpy as np
from .Abstract_Ranker import Abstract_Regularizer
import torch.nn.functional as F
import torch
r"""
Reg
################################################
Reference: Toshihiro Kamishima and Shotaro Akaho. 2017. Considerations on recommendation independence for a fnd-good-items task

##Note that in this code, we extend the two-group cases into multi-group cases

"""


class Reg(Abstract_Regularizer):
    def __init__(self, config, group_weight):
        super().__init__(config)


    def fairness_loss(self, input_dict):
        losses = input_dict['scores'] #[B]
        return torch.var(losses)



